#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Utility functions for downloading datasets from OSF
"""
import os
import re
from collections import defaultdict
from pkg_resources import resource_filename
import json

MATCH = re.compile(
    r'space-(\S+)_(?:den|res)-(\d+[k|m]{1,2})_'
)

def _groupby_match(fnames):
    """"
    Groups files in `fnames` by (space, res/den)

    Parameters
    ----------
    fnames : list-of-str
        Filenames to be grouped
    return_single : bool, optional
        If there is only group of filenames return a list instead of a dict.
        Default: False

    Returns
    -------
    groups : dict-of-str
        Where keys are tuple (source, desc, space, res/den) and values are
        lists of filenames
    """

    out = defaultdict(list)
    for fn in fnames:
        out[MATCH.search(fn).groups()].append(fn)

    out = {k: v if len(v) > 1 else v[0] for k, v in out.items()}
    
    if len(out) == 1:
        out = list(out.values())[0]

    return out

def _osfify_urls(data):
    """
    Formats `data` object with OSF API URL

    Parameters
    ----------
    data : object
        If dict with a `url` key, will format OSF_API with relevant values
    return_restricted : bool, optional
        Whether to return restricted annotations. These will only be accesible
        with a valid OSF token. Default: True

    Returns
    -------
    data : object
        Input data with all `url` dict keys formatted
    """

    OSF_API = "https://files.osf.io/v1/resources/{}/providers/osfstorage/{}"

    if isinstance(data, str) or data is None:
        return data
    elif 'url' in data:
        # if url is None then we this is a malformed entry and we should ignore
        if data['url'] is None:
            return
        # if the url isn't a string assume we're supposed to format it
        elif not isinstance(data['url'], str):
            data['url'] = OSF_API.format(*data['url'])

    try:
        for key, value in data.items():
            data[key] = _osfify_urls(value)
    except AttributeError:
        for n, value in enumerate(data):
            data[n] = _osfify_urls(value)
        # drop the invalid entries
        data = [d for d in data if d is not None]

    return data

def get_data_dir(data_dir=None):
    """
    Gets path to eigenstrapping directory

    Parameters
    ----------
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'EIGEN_DATA'; if that is not set, will
        use `~/neuromaps-data` instead. Default: None

    Returns
    -------
    data_dir : str
        Path to use as data directory
    """

    if data_dir is None:
        data_dir = os.environ.get('EIGEN_DATA',
                                  os.path.join('~', 'eigenstrapping-data'))
    data_dir = os.path.expanduser(data_dir)
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)

    return data_dir

def _match_files(info, **kwargs):
    """
    Matches datasets in `info` to relevant keys

    Parameters
    ----------
    info : list-of-dict
        Information on files
    kwargs : key-value pairs
        Values of data in `info` on which to match

    Returns
    -------
    matched : list-of-dict
        Annotations with specified values for keys
    """

    # tags should always be a list
    tags = kwargs.get('tags')
    if tags is not None and isinstance(tags, str):
        kwargs['tags'] = [tags]
        
    # 'den' and 'res' are a special case because these are mutually exclusive
    # values (only one will ever be set for a given annotation) so we want to
    # match on _either_, not both, if and only if both are provided as keys.
    # if only one is specified as a key then we should exclude the other!
    denres = []
    for vals in (kwargs.get('den'), kwargs.get('res')):
        vals = [vals] if isinstance(vals, str) else vals
        if vals is not None:
            denres.extend(vals)

    out = []
    for dset in info:
        match = True
        for key in ('space', 'hemi', 'tags', 'format'):
            comp, value = dset.get(key), kwargs.get(key)
            if value is None:
                continue
            elif value is not None and comp is None:
                match = False
            elif isinstance(value, str):
                if value != 'all':
                    match = match and comp == value
            else:
                func = all if key == 'tags' else any
                match = match and func(f in comp for f in value)
        if len(denres) > 0:
            match = match and (dset.get('den') or dset.get('res')) in denres
        if match:
            out.append(dset)

    return out

def get_dataset_info(name):
    """
    Returns information for requested dataset `name`

    Parameters
    ----------
    name : str
        Name of dataset

    Returns
    -------
    dataset : dict or list-of-dict
        Information on requested data
    """

    fn = resource_filename('eigenstrapping',
                           os.path.join('datasets', 'osf.json'))
    with open(fn) as src:
        osf_resources = _osfify_urls(json.load(src))

    try:
        resource = osf_resources[name]
    except KeyError:
        raise KeyError("Provided dataset '{}' is not valid. Must be one of: {}"
                       .format(name, sorted(osf_resources.keys())))

    return resource


# -*- coding: utf-8 -*-
"""
Functions for calculating and manipulating spatial autocorrelation
Author: Ross Markello, netneurolab
"""

import os
import tempfile
from pathlib import Path
import matplotlib.pyplot as plt

import nibabel as nib
import numpy as np
from scipy import fftpack, stats as sstats
from scipy.spatial import cKDTree
from scipy.optimize import minimize
from scipy.special import erfc

from netneurotools.datasets import make_correlated_xy
from netneurotools.freesurfer import check_fs_subjid
from netneurotools.utils import run
from neuromaps import transforms, datasets

from brainspace import mesh

VOL2SURF = 'mri_vol2surf --src {} --out {} --hemi {} --mni152reg ' \
           '--trgsubject fsaverage5 --projfrac 0.5 --interp nearest'
MSEED = 4294967295

def estimate_ex_params(x, initial_params=[0., 1., 1.], plot=False):
    """
    Estimate the parameters of an ex-Gaussian distribution from given `data`
    where the PDF of the distribution is
    
                          tau       tau
    p(x|mu, sigma, tau) = --- exp ( --- (2 * mu + tau * sigma**2 - 2 * x) ) *
                           2         2
                                  mu + tau * sigma ** 2 - x
                          erfc( -------------------------- ),
                                      sqrt(2) * sigma
                        
                        where erfc = 1 - erf.
                        
    Uses negative log-likelihood minimization.
    
    Parameters
    ----------
    x : np.ndarray of shape (N,)
        Data array to estimate parameters, must be 1D
    initial_params : list or tuple, optional
        Initial guesses for parameters mu, sigma, and tau. Default is 
        [0., 1., 1.].
    plot : bool, optional
        Flag to plot data with overlay curve estimated by parameters. Default 
        is False.

    Returns
    -------
    np.ndarray of shape (3,)
        Parameter estimates for mu, sigma, and tau.

    """
    
    # Mask invalid values
    masked_data = np.ma.masked_invalid(x)
    
    result = minimize(_neg_log_likelihood,
                      x0=initial_params,
                      method='Nelder-Mead',
                      args=(masked_data)
            )
    
    mu, sigma, tau = result.x
    
    if plot:
        x_values = np.linspace(np.min(masked_data), np.max(masked_data), 1000)
        pdf = ex_gaussian_pdf(x_values, mu, sigma, tau)
        fig = plt.figure(figsize=(10, 6), constrained_layout=False)
        ax = fig.add_axes([0.12, 0.15, 0.8, 0.77])
        
        ax.plot(x_values, pdf, label=f'ex-Gaussian PDF ($\mu$={mu:.3f},\n$\sigma$={sigma:.3f}, $\tau$={tau:.3f})', color='blue')
        ax.hist(masked_data.compressed(), bins=50, density=True, alpha=0.6, color='gray', label='Data')
        ax.set_xlabel('Value', fontsize=20)
        ax.set_ylabel('Density', fontsize=20)
        leg = ax.legend(loc=0, fontsize=20)
        leg.get_frame().set_linewidth(0.0)
        plt.show()        
    
    return np.array((mu, sigma, tau))
    
def _neg_log_likelihood(params, data):
    mu, sigma, tau = params
    return -np.sum(np.ma.log(ex_gaussian_pdf(data, mu, sigma, tau)))    

def ex_gaussian_pdf(x, mu, sigma, tau):
    return (tau /2) * np.exp((tau / 2) * (2 * mu + tau * sigma**2 - 2 * x)) * erfc((mu + tau * sigma**2 - x) / (np.sqrt(2) * sigma))

def bump_function(coords, vertex, radius=4., value=1.):
    """
    Given a list of 3D coordinates, this bump function sets the value at 
    `vertex` in a 1D array to `value`. It also sets the value in the 1D array 
    to `value` for any vertex that is within the specified `radius` from the 
    vertex at initial `vertex`.

    Parameters
    ----------
    coords : np.ndarray of shape (N, 3)
        List of 3D coordinates
    vertex : int
        Index of centroid of bump function
    distance : float, optional
        Radius of bump function. Default is 4
    value : float, optional
        Value to substitute for bump function

    Returns
    -------
    bump_array : np.ndarray of shape (N,)
        Output bump function.

    """
    
    num_vertices = coords.shape[1]
    bump_array = np.zeros(num_vertices)
    
    # set initial vertex to value
    bump_array[vertex] = value
    vertex_coords = coords[vertex]
    
    for i, vert in enumerate(coords):
        # Calculate the Euclidean distance between the two points
        dist = np.linalg.norm(vertex_coords - vert)
        if dist <= radius:
            bump_array[i] = value
    
    return bump_array

def _objective(alpha, x, y_obs):
    y_pred = x ** -alpha
    res = y_obs - y_pred
    return np.sum(res**2)

def fit_alpha(p, coeffs, fmin=5.):
    """
    Fit power spectrum with a model specified in `model` to estimate
    exponent of spatial autocorrelation (by Wiener-Khinchin theorem).
    Minimizes the objective function (sum of squared residuals).

    Parameters
    ----------
    p : np.ndarray of shape (num_modes,)
        Modal power spectrum
    coeffs : np.ndarray of shape (num_modes,)
        Mode coefficients
    model : method
        Function to fit data. e.g., np.exp(-alpha * p)
    fmin : float
        Minimum spatial frequency to fit (in mm). Default 5.0

    Returns
    -------
    est_alpha : float
        Estimated alpha
    cov : np.ndarray
        Covariance for goodness of fit

    """
    # spatial frequencies
    num_modes = p.shape[0]
    x = np.linspace(fmin, np.max(np.abs(coeffs)), num_modes)
    
    params = minimize(_objective, x0=[1.0], args=(x, p))
    est_alpha = params.x[0]
    
    return est_alpha
    

def map_data_to_grid(points, faces, data, mask=None, spacing=0.5, k=1, 
                     avg_method='mean', eps=0, n_jobs=1):
    """
    Map data from a (cortical or subcortical) mesh of elements ``points``
    and ``faces`` to a square 2D matrix. Uses k-nearest neighbors tree query
    approach.

    Parameters
    ----------
    points : np.ndarray of shape (N,3)
        Mesh coordinates of ``data``
    faces : np.ndarray of shape (F,3)
        Array of face connectivity of mesh
    data : np.ndarray of shape (N,) or (N,M)
        Data to resample to grid, must have the same number of rows as
        ``points`` but can have arbitrarily many columns (data points)
    mask : np.ndarray of bool of shape (N,)
        Mask of surface mesh to exclude from final resampling. Must be an array
        of np.ndarray of bool of True (to keep) and False (to exclude) for each
        index in ``points``
    spacing : non-negative float, optional
        Spacing in mm of grid to resample data. Determines density of resulting
        grid.
    k : int, optional
        Number of neighbors to resample. Default 1
    avg_method : str, optional
        If k is larger than 1, use ``avg_method`` to resample reweighted data
        to grid. Accepts "mean", "weighted_mean", "mode", or "weighted_mode".
    eps : non-negative float, optional
        Error tolerance between k nearest distances on grid and points. Default 0
    n_jobs : int, optional
        Number of workers for tree search. Default 1        

    Returns
    -------
    grid_values : np.ndarray of shape (size, size) if data was vector or
        np.ndarray of shape (M, size, size) if data was array
        
        Interpolated values of ``data`` on 2D grid for each column in ``data``

    """
    m = mesh.mesh_creation.build_polydata(points, cells=faces)
    
    if mask is not None:
        masked = mesh.mesh_operations.mask_points(m, mask)
        data = mesh.array_operations.resample_pointdata(m, masked, data=data)
        m = masked
    
    # get boundary
    bp = mesh.mesh_elements.get_boundary_points(m)
    
    # use min and max to define grid
    min_x, max_x = np.min(points[:, 0]), np.max(points[:, 0])
    min_y, max_y = np.min(points[:, 1]), np.max(points[:, 1])
    x = np.arange(min_x, max_x, spacing)
    y = np.arange(min_y, max_y, spacing)
    grid_x, grid_y = np.meshgrid(x, y)
    
    grid_points = np.c_[grid_x.ravel(), grid_y.ravel()]
    
    # use query tree to find k-nearest neighbors
    tree = cKDTree(points, leafsize=20, compact_nodes=False,
                   copy_data=False, balanced_tree=False)
    d, idx = tree.query(grid_points, k=k, eps=eps, workers=n_jobs,
                        distance_upper_bound=eps+np.finfo(np.float64).eps)

    weights = 1.0 / d
    
    new_data = np.zeros(data.shape)
    
    for i in range(data.shape[1]):
        feat = data[:, i]
        
        if k == 1:
            f = feat[idx]
        elif avg_method == 'mean':
            f = np.nanmean(feat[idx], axis=1)
        elif avg_method == 'weighted_mean':
            f = np.average(feat[idx], weights=weights, axis=1)
        elif avg_method == 'mode':
            f = sstats.mode(feat[idx], axis=1)[0].squeeze()
        elif avg_method == 'weighted_mode':
            f = sstats.weighted_mode(feat[idx], weights=weights, axis=1)[0].squeeze()
            f = f.astype(data.dtype)
        else:
            raise ValueError('Unknown avg_method: {0}'.format(str(avg_method)))
           
    if data.ndim > 1:
        grid_values = new_data.reshape(i, grid_x.shape[0], grid_x.shape[1])
        return grid_values
    
    grid_values = new_data.reshape(grid_x.shape)
    
    return grid_values

def morans_i(dist, y, normalize=False, local=False, invert_dist=True):
    """
    Calculates Moran's I from distance matrix `dist` and brain map `y`

    Parameters
    ----------
    dist : (N, N) array_like
        Distance matrix between `N` regions / vertices / voxels / whatever
    y : (N,) array_like
        Brain map variable of interest
    normalize : bool, optional
        Whether to normalize rows of distance matrix prior to calculation.
        Default: False
    local : bool, optional
        Whether to calculate local Moran's I instead of global. Default: False
    invert_dist : bool, optional
        Whether to invert the distance matrix to generate a weight matrix.
        Default: True

    Returns
    -------
    i : float
        Moran's I, measure of spatial autocorrelation
    """

    # convert distance matrix to weights
    if invert_dist:
        with np.errstate(divide='ignore'):
            dist = 1 / dist
    np.fill_diagonal(dist, 0)

    # normalize rows, if desired
    if normalize:
        dist /= dist.sum(axis=-1, keepdims=True)

    # calculate Moran's I
    z = y - y.mean()
    if local:
        with np.errstate(all='ignore'):
            z /= y.std()

    zl = np.squeeze(dist @ z[:, None])
    den = (z * z).sum()

    if local:
        return (len(y) - 1) * z * zl / den

    return len(y) / dist.sum() * (z * zl).sum() / den


def _fftind(x, y, z):
    """
    Return 3D shifted Fourier coordinates

    Returned coordinates are shifted such that zero-frequency component of the
    square grid with shape (x, y, z) is at the center of the spectrum

    Parameters
    ----------
    {x,y,z} : int
        Size of array to be generated

    Returns
    -------
    k_ind : (3, x, y, z) np.ndarray
        Shifted Fourier coordinates, where:
            k_ind[0] : k_x components
            k_ind[1] : k_y components
            k_ind[2] : k_z components

    Notes
    -----
    See scipy.fftpack.fftshift

    References
    ----------
    Based on code from Burt et al., 2020, NeuroImage
    """

    k_ind = np.mgrid[:x, :y, :z]
    zero = np.array([int((n + 1) / 2) for n in [x, y, z]])
    while zero.ndim < k_ind.ndim:
        zero = np.expand_dims(zero, -1)
    k_ind = fftpack.fftshift(k_ind - zero)

    return k_ind


def gaussian_random_field(x, y, z, noise=None, alpha=3.0, normalize=True,
                          seed=None):
    """
    Generate a Gaussian random field with k-space power law |k|^(-alpha/2).

    Parameters
    ----------
    {x,y,z} : int
        Grid size of generated field
    noise : (x, y, z) array_like, optional
        Noise array to which gaussian smoothing is added. If not provided an
        array will be created by drawing from the standard normal distribution.
        Default: None
    alpha : float (positive), optional
        Power (exponent) of the power-law distribution. Default: 3.0
    normalize : bool, optional
        Normalize the returned field to unit variance. Default: True
    seed : None, int, default_rng, optional
        Random state to seed `noise` generation. Default: None

    Returns
    -------
    gfield : (x, y, z) np.ndarray
        Realization of Gaussian random field

    References
    ----------
    Based on code from Burt et al., 2020, NeuroImage
    """

    rs = np.random.default_rng(seed)

    if not alpha:
        return rs.normal(size=(x, y, z))

    assert alpha > 0

    # k-space indices
    k_idx = _fftind(x, y, z)

    # define k-space amplitude as a power law 1/|k|^(alpha/2)
    amplitude = np.power(np.sum([k ** 2 for k in k_idx], axis=0) + 1e-10,
                         -alpha / 2.0)
    amplitude[0, 0, 0] = 0  # remove zero-freq mean shit

    # generate a complex gaussian random field where phi = phi_1 + i*phi_2
    if noise is None:
        noise = rs.normal(size=(x, y, z))
    elif noise.shape != (x, y, z):
        try:
            noise = noise.reshape(x, y, z)
        except ValueError:
            raise ValueError('Provided noise cannot be reshape to target: '
                             f'({x}, {y}, {z})')

    # transform back to real space
    gfield = np.fft.ifftn(np.fft.fftn(noise) * amplitude).real

    if normalize:
        return (gfield - gfield.mean()) / gfield.std()

    return gfield


def make_tmpname(suffix, prefix=None):
    """
    Stupid helper function because :man_shrugging:

    Parameters
    ----------
    suffix : str
        Suffix of created filename

    Returns
    -------
    fn : str
        Temporary filename; user is responsible for deletion
    """

    # I don't want to deal with a bunch of nested tempfile.NameTemporaryFile
    # in the create_surface_grf() function so this is the easiest way to do
    # things that's safe from race conditions :man_shrugging:

    fd, fn = tempfile.mkstemp(suffix=suffix, prefix=prefix)
    os.close(fd)

    return fn

def make_decimated_surfaces(initial_surface, per=0.1, medial=None, 
                            outdir="datasets/decimated", prefix='decsurf', n_jobs=1):
    """
    Creates decimated surfaces based on ``initial_mesh`` using brainspace.

    Parameters
    ----------
    initial_surface : str
        Path to surface to decimate.
    medial : str
        Path to medial wall mask of .txt, where 1 corresponds to non-medial wall
        and 0 corresponds to medial wall.
    per : float, optional
        "Percentage" of vertices to remove each step, where 0.0 < ``per`` < 1.0.
        The default is 0.1.
    outdir : str, optional
        Output directory for surfaces. Default is "datasets/decimated"
    prefix : str, optional
        Prefix of the decimated surfaces. Default is "decsurf".
    n_jobs : int, optional
        Number of workers to use for decimation. Default is 1.

    Returns
    -------
    surface_fnames : list of str
        List of absolute paths of decimated surfaces
    medial_fnames : list of str
        List of absolute paths of resampled medial wall masks 
        (if ``medial`` is given)

    """
    
    outdir = Path(outdir)
    prefix = str(prefix)
    surface = mesh.mesh_io.read_surface(initial_surface)
    surface_fnames = []
    if medial:
        medial = np.loadtxt(medial)
        medial_fnames = []
    
    if per <= 0.0 or per >= 1.0:
        raise ValueError('Percentage of vertices to remove must be between 0 and 1')
    
    n_steps = np.int(np.floor(1.0 / per))
    
    for step in range(n_steps):
        factor = 1.0 - per*step
        if factor == 1.0:
            n_vertices = surface.GetNumberOfPoints()
            fn = os.path.join(outdir, f'{prefix}_verts={n_vertices}.gii')
            surface_fnames.append(os.path.abspath(fn))
            mesh.mesh_io.write_surface(surface, fn, otype='gii')
            if medial is not None:
                medial_fn = os.path.join(outdir, f'{prefix}_medial_verts={n_vertices}.txt')
                np.savetxt(medial_fn, medial, fmt='%d')
                medial_fnames.append(os.path.abspath(medial_fn))
            continue
                
        # mask surface
        new_surface_mask = mesh.mesh_cluster.sample_points_decimation(surface,
                                keep=factor, n_jobs=n_jobs)
        new_surface = mesh.mesh_operations.mask_points(surface, new_surface_mask)
        n_vertices = new_surface.GetNumberOfPoints()
        fn = os.path.join(outdir, f'{prefix}_verts={n_vertices}.gii')
        surface_fnames.append(os.path.abspath(fn))
        mesh.mesh_io.write_surface(new_surface, fn, otype='gii')
        
        if medial is not None:
            medial_out = mesh.array_operations.resample_pointdata(surface, new_surface,
                                data=medial, append=False)
            medial_fn = os.path.join(outdir, f'{prefix}_medial_verts={n_vertices}.txt')
            np.savetxt(medial_fn, medial_out, fmt='%d')
            medial_fnames.append(os.path.abspath(medial_fn))
            
    
    if medial is not None:
        return surface_fnames, medial_fnames
    
    return surface_fnames

def make_correlated_exgaussian(corr=0.15, size=10000, seed=None, tol=0.005, mu=0, sigma=1, tau=0.3):
    """
    Generates two exGaussian random vectors that are correlated at approximately the specified correlation value.

    Parameters
    ----------
    corr : float, optional
        The approximate correlation desired between the two generated exGaussian vectors. Default: 0.15
    size : int, optional
        Desired size of the generated vectors. Default: 10000
    seed : int or np.random.RandomState instance or None, optional
        Seed for random number generation. Default: None
    tol : float, optional
        Tolerance of correlation between generated vectors and specified `corr`. Default: 0.005
    mu : float, optional
        Mean of the Gaussian component for the exGaussian distribution. Default: 0
    sigma : float, optional
        Standard deviation of the Gaussian component for the exGaussian distribution. Default: 1
    tau : float, optional
        Mean and standard deviation of the exponential component for the exGaussian distribution. Default: 0.3

    Returns
    -------
    x_exgaussian, y_exgaussian : numpy.ndarray
        Two random vectors of size `size` with exGaussian distributions and correlation specified by `corr`.

    Examples
    --------
    >>> x, y = make_correlated_exgaussian(corr=0.2, seed=42)
    >>> np.corrcoef(x, y)[0, 1]
    0.2003...
    """
    
    # Generate two correlated vectors using make_correlated_xy
    rs = np.random.RandomState(seed)
    x, y = make_correlated_xy(corr=corr, size=size, seed=seed, tol=tol)
    
    # Transform each vector from normal to exgaussian using quantile transformation
    exgauss_samples = [exgaussian(rs, mu, sigma, tau) for _ in range(size)]
    exgauss_samples.sort()
    
    x_exgaussian = [exgauss_samples[int(rank)] for rank in sstats.rankdata(x) - 1]
    y_exgaussian = [exgauss_samples[int(rank)] for rank in sstats.rankdata(y) - 1]
    
    return np.array(x_exgaussian), np.array(y_exgaussian)

def exgaussian(rs, mu=0, sigma=1, tau=0.3):
    "Generate an exGaussian random variable"
    gaussian_component = rs.normal(mu, sigma)
    exponential_component = rs.exponential(tau)
    return gaussian_component + exponential_component

def exgaussian_field(x, y, z, alpha, noise=None, normalize=True, seed=None, 
                     mu=0., sigma=1., tau=0.3):
    """
    Generate a ex-Gaussian random field with k-space power law |k|^(-alpha/2).

    Parameters
    ----------
    {x,y,z} : int
        Grid size of generated field
    noise : (x, y, z) array_like, optional
        Noise array to which gaussian smoothing is added. If not provided an
        array will be created by drawing from the standard normal distribution.
        Default: None
    alpha : float (positive), optional
        Power (exponent) of the power-law distribution. Default: 3.0
    normalize : bool, optional
        Normalize the returned field to unit variance. Default: True
    seed : None, int, default_rng, optional
        Random state to seed `noise` generation. Default: None

    Returns
    -------
    gfield : (x, y, z) np.ndarray
        Realization of Gaussian random field

    References
    ----------
    Based on code from Burt et al., 2020, NeuroImage
    """
    
    rs = np.random.default_rng(seed)
    # k-space indices
    k_idx = _fftind(x, y, z)

    # define k-space amplitude as a power law 1/|k|^(alpha/2)
    amplitude = np.power(np.sum([k ** 2 for k in k_idx], axis=0) + 1e-10,
                         -alpha / 2.0)
    amplitude[0, 0, 0] = 0  # remove zero-freq mean shift

    # generate a complex ex-Gaussian random field where phi = phi_1 + i*phi_2
    if noise is None:
        noise = np.array([exgaussian(rs, mu=mu, sigma=sigma, tau=tau) for _ in range(x*y*z)]).reshape(x, y, z)
    elif noise.shape != (x, y, z):
        try:
            noise = noise.reshape(x, y, z)
        except ValueError:
            raise ValueError('Provided noise cannot be reshaped to target: '
                             f'({x}, {y}, {z})')

    # transform back to real space
    gfield = np.fft.ifftn(np.fft.fftn(noise) * amplitude).real
    
    if normalize:
        return (gfield - gfield.mean()) / gfield.std()

    return gfield

def create_surface_exgaussian(noise=None, alpha=3.0, normalize=True, seed=None,
                       medial_val=None, mu=0., sigma=1., tau=0.3, hemis=['lh']):
    """
    Generates ex-GRF on surface (fsaverage5)

    Uses exgaussian_field() and mri_vol2surf to generate GRF

    Parameters
    ----------
    noise : (x, y, z) array_like, optional
        Noise array to which gaussian smoothing is added. If not provided an
        array will be created by drawing from the standard normal distribution.
        Default: None
    alpha : float (positive), optional
        Power (exponent) of the power-law distribution. Default: 3.0
    normalize : bool, optional
        Normalize the returned field to unit variance. Default: True
    seed : None, int, default_rng, optional
        Random state to seed GRF. Default: None

    Returns
    -------
    data : (20484,) np.ndarray
        Surface representation of GRF
    """

    affine = np.eye(4) * 2
    affine[:, -1] = [-90, -90, -72, 1]

    gfield = exgaussian_field(91, 109, 91, noise=noise, alpha=alpha,
                                   normalize=normalize, seed=seed,
                                   mu=mu, sigma=sigma, tau=tau)
    fn = make_tmpname(suffix='.nii.gz')
    nib.save(nib.nifti1.Nifti1Image(gfield, affine), fn)
    
    if len(hemis) < 2:
        outname = make_tmpname(suffix='.mgh')
        run(VOL2SURF.format(fn, outname, hemis[0]), quiet=True)
        data = nib.load(outname).get_fdata().squeeze()
        os.remove(outname)
    else:
        data = np.zeros((20484,))
        for n, hemi in enumerate(hemis):
            outname = make_tmpname(suffix='.mgh')
            run(VOL2SURF.format(fn, outname, hemi), quiet=True)
            sl = slice(len(data) // 2 * n, len(data) // 2 * (n + 1))
            data[sl] = nib.load(outname).get_fdata().squeeze()
            os.remove(outname)

    os.remove(fn)

    if medial_val is not None:
        data = _mod_medial(_mod_medial(data, True), False, medial_val)

    return data

def create_surface_grf(noise=None, alpha=3.0, normalize=True, seed=None,
                       medial_val=None):
    """
    Generates GRF on surface (fsaverage5)

    Uses gaussian_random_field() and mri_vol2surf to generate GRF

    Parameters
    ----------
    noise : (x, y, z) array_like, optional
        Noise array to which gaussian smoothing is added. If not provided an
        array will be created by drawing from the standard normal distribution.
        Default: None
    alpha : float (positive), optional
        Power (exponent) of the power-law distribution. Default: 3.0
    normalize : bool, optional
        Normalize the returned field to unit variance. Default: True
    seed : None, int, default_rng, optional
        Random state to seed GRF. Default: None

    Returns
    -------
    data : (20484,) np.ndarray
        Surface representation of GRF
    """

    affine = np.eye(4) * 2
    affine[:, -1] = [-90, -90, -72, 1]

    gfield = gaussian_random_field(91, 109, 91, noise=noise, alpha=alpha,
                                   normalize=normalize, seed=seed)
    fn = make_tmpname(suffix='.nii.gz')
    nib.save(nib.nifti1.Nifti1Image(gfield, affine), fn)

    data = np.zeros((20484,))
    for n, hemi in enumerate(('lh', 'rh')):
        outname = make_tmpname(suffix='.mgh')
        run(VOL2SURF.format(fn, outname, hemi), quiet=True)
        sl = slice(len(data) // 2 * n, len(data) // 2 * (n + 1))
        data[sl] = nib.load(outname).get_fdata().squeeze()
        os.remove(outname)

    os.remove(fn)

    if medial_val is not None:
        data = _mod_medial(_mod_medial(data, True), False, medial_val)

    return data

def create_decimated_grf(surface, hemi='L', noise=None, alpha=3.0, normalize=True, seed=None,
                         medial=None):
    """
    Generates GRF on surface given by brainspace decimation. (1) resamples
    from MNI152 2mm space to fsLR 164k standard space, then (2) resamples GRF 
    to decimated surface in ``surface``. Makes medial wall vertices in ``data`` 
    equal to NaN if not None.

    Parameters
    ----------
    surface : str
        Decimated surface, must be single hemisphere
    hemi : str
        Hemisphere that decimated surface is in. Must be 'L' or 'R', default is 'L'
    noise : (x, y, z) array_like, optional
        Noise array to which gaussian smoothing is added. If not provided an
        array will be created by drawing from the standard normal distribution.
        Default: None
    alpha : float (positive), optional
        Power (exponent) of the power-law distribution. Default: 3.0
    normalize : bool, optional
        Normalize the returned field to unit variance. Default: True
    seed : None, int, default_rng, optional
        Random state to seed GRF. Default: None
    medial : str
        Medial wall mask where medial wall is 0 and everything else is 1

    Returns
    -------
    data : (n_vertices,) np.ndarray
        Surface representation of GRF where ``n_vertices`` is the number of
        vertices in the input ``surface``. If ``medial`` is given, np.nan values
        are inserted into corresponding indices.

    """
    
    affine = np.eye(4) * 2
    affine[:, -1] = [-90, -90, -72, 1]
    
    gfield = gaussian_random_field(91, 109, 91, noise=noise, alpha=alpha,
                                   normalize=normalize, seed=seed)
    fn = make_tmpname(suffix='.nii.gz')
    nib.save(nib.nifti1.Nifti1Image(gfield, affine), fn)
    
    surface = mesh.mesh_io.read_surface(surface)
    
    if hemi == 'L':
        gfield = transforms.mni152_to_fslr(fn, fslr_density='164k')[0].agg_data()
        standard = str(datasets.fetch_fslr(density='164k', verbose=0)['midthickness'][0])
    elif hemi == 'R':
        gfield = transforms.mni152_to_fslr(fn, fslr_density='164k')[1].agg_data()
        standard = str(datasets.fetch_fslr(density='164k', verbose=0)['midthickness'][1])
    else:
        raise ValueError("Hemi must be 'L' or 'R'")
    
    standard = mesh.mesh_io.read_surface(standard)
    data = mesh.array_operations.resample_pointdata(surface, standard,
                        data=gfield, append=False)
    
    if medial:
        medial = np.loadtxt(medial).astype(np.bool_)
        medial_mask = np.logical_not(medial)
        data[medial_mask] = np.nan
    
    os.remove(fn)
    
    return data


def matching_multinorm_decimated_grfs(surface, corr, hemi='L', medial=None, 
                                      tol=0.005, alpha=3.0, normalize=True, seed=None):
    """
    Generates two surface GRFs that correlate at r = `corr`

    Starts by generating two random variables from a multivariate normal
    distribution with correlation `corr`, adds spatial autocorrelation with
    specified `alpha`, and projects to the surface. Continues this procedure
    until two variables are generated that have correlation `corr` on the
    surface.

    Parameters
    ----------
    surface : str
        Decimated surface, must be single hemisphere
    corr : float
        Desired correlation of generated GRFs
    hemi : str
        Hemisphere that decimated surface is in. Must be 'L' or 'R', default is 'L'
    tol : float
        Tolerance for correlation between generated GRFs
    alpha : float (positive), optional
        Exponent of the power-law distribution. Only used if `use_gstools` is
        set to False. Default: 3.0
    normalize : bool, optional
        Whether to normalize the returned field to unit variance. Default: True
    seed : None, int, default_rng, optional
        Random state to seed GRF generation. Default: None

    Returns
    -------
    xs, ys : (n_vertices)

    """
    
    rs = np.random.default_rng(seed)
    acorr = np.inf
    
    surface_mesh = mesh.mesh_io.read_surface(surface)
    
    while np.abs(np.abs(acorr) - corr) > tol:
        if alpha > 0:
            x, y = make_correlated_xy(corr, size=902629,
                                      seed=rs.integers(MSEED))
            # smooth and project
            xs = create_decimated_grf(surface, hemi=hemi, noise=x, 
                                      alpha=alpha, normalize=normalize,
                                      medial=medial)
            ys = create_decimated_grf(surface, hemi=hemi, noise=y, 
                                      alpha=alpha, normalize=normalize,
                                      medial=medial)
            
            if medial is not None:
                # check if removing medial wall has altered correlation or if
                # x and y remain sufficiently correlated
                acorr = np.corrcoef(xs[np.logical_not(np.isnan(xs))], ys[np.logical_not(np.isnan(ys))])[0, 1]
            else:
                acorr = np.corrcoef(xs, ys)[0, 1]
            
        else:
            xs, ys = make_correlated_xy(corr, size=surface_mesh.GetNumberOfPoints(),
                                        seed=rs.integers(MSEED))
    
    if acorr < 0:
        ys *= -1
        
    if normalize:
        xs, ys = sstats.zscore(xs), sstats.zscore(ys)
        
    return xs, ys

def _mod_medial(data, remove=True, val=0):
    """
    Removes (inserts) medial wall from (into) `data` from fsaverage5 surface

    Parameters
    ----------
    data : (20484,) array_like
        Surface data
    remove : bool, optional
        Whether to remove medial wall instead of inserting it. Assumes input
        has (does not have) medial wall. Default: True
    val : float, optional
        What value to insert if `remove=False`. Default: 0

    Returns
    -------
    out : np.ndarray
        Provided surface `data` with medial wall removed/inserted
    """
    if len(data) <= 10242:
        subj, path = check_fs_subjid('fsaverage5')
        lh = [
            nib.freesurfer.read_label(
                os.path.join(path, subj, 'label', f'{h}.Medial_wall.label')
            )
            for h in (['lh'])
        ]
        lhm = np.ones(10242, dtype=bool)
        lhm[lh] = False
    
        if remove:
            return data[lhm]
        else:
            xd = np.ones(10242) * val
            xd[lhm] = data
            return xd
    else:
        subj, path = check_fs_subjid('fsaverage5')
        lh, rh = [
            nib.freesurfer.read_label(
                os.path.join(path, subj, 'label', f'{h}.Medial_wall.label')
            )
            for h in ('lh', 'rh')
        ]
        lhm, rhm = np.ones(10242, dtype=bool), np.ones(10242, dtype=bool)
        lhm[lh], rhm[rh] = False, False
    
        if remove:
            x, y = np.split(data, 2)
            return np.hstack((x[lhm], y[rhm]))
        else:
            x, y = np.split(data, [np.sum(lhm)])
            xd, yd = np.ones(10242) * val, np.ones(10242) * val
            xd[lhm], yd[rhm] = x, y
            return np.hstack((xd, yd))


def matching_multinorm_grfs(corr, tol=0.005, *, alpha=3.0, normalize=True,
                            seed=None, debug=False):
    """
    Generates two surface GRFs (fsaverage5) that correlate at r = `corr`

    Starts by generating two random variables from a multivariate normal
    distribution with correlation `corr`, adds spatial autocorrelation with
    specified `alpha`, and projects to the surface. Continues this procedure
    until two variables are generated that have correlation `corr` on the
    surface.

    Parameters
    ----------
    corr : float
        Desired correlation of generated GRFs
    tol : float
        Tolerance for correlation between generated GRFs
    alpha : float (positive), optional
        Exponent of the power-law distribution. Only used if `use_gstools` is
        set to False. Default: 3.0
    normalize : bool, optional
        Whether to normalize the returned field to unit variance. Default: True
    seed : None, int, default_rng, optional
        Random state to seed GRF generation. Default: None
    debug : bool, optional
        Whether to print debug info

    Return
    ------
    x, y : (20484,) np.ndarray
        Generated surface GRFs
    """

    rs = np.random.default_rng(seed)

    acorr, n = np.inf, 0
    while np.abs(np.abs(acorr) - corr) > tol:
        if alpha > 0:
            x, y = make_correlated_xy(corr, size=902629,
                                      seed=rs.integers(MSEED))
            # smooth correlated noise vectors + project to surface
            xs = create_surface_grf(noise=x, alpha=alpha, normalize=normalize)
            ys = create_surface_grf(noise=y, alpha=alpha, normalize=normalize)
        else:
            xs, ys = make_correlated_xy(corr, size=20484,
                                        seed=rs.integers(MSEED))

        # remove medial wall to ensure data are still sufficiently correlated.
        # this is important for parcellations that will ignore the medial wall
        xs, ys = _mod_medial(xs, remove=True), _mod_medial(ys, remove=True)
        acorr = np.corrcoef(xs, ys)[0, 1]

        if debug:
            # n:>3 because dear lord i hope it doesn't take more than 999 tries
            print(f'{n:>3}: {acorr:>6.3f}')
            n += 1

    if acorr < 0:
        ys *= -1

    if normalize:
        xs, ys = sstats.zscore(xs), sstats.zscore(ys)

    return _mod_medial(xs, remove=False), _mod_medial(ys, remove=False)

def matching_multinorm_exgrfs(corr, tol=0.005, alpha=3.0, params=[0., 1., 1.], normalize=True, seed=None, debug=False):
    """
    Generates two surface ex-GRFs (fsaverage5) that correlate at r = `corr`

    Starts by generating two random variables from a multivariate normal
    distribution with correlation `corr`, adds spatial autocorrelation with
    specified `alpha`, and projects to the surface. Continues this procedure
    until two variables are generated that have correlation `corr` on the
    surface.

    Parameters
    ----------
    corr : float
        Desired correlation of generated GRFs
    tol : float
        Tolerance for correlation between generated GRFs
    alpha : float (positive), optional
        Exponent of the power-law distribution. Only used if `use_gstools` is
        set to False. Default: 3.0
    normalize : bool, optional
        Whether to normalize the returned field to unit variance. Default: True
    seed : None, int, default_rng, optional
        Random state to seed GRF generation. Default: None
    debug : bool, optional
        Whether to print debug info

    Return
    ------
    x, y : (20484,) np.ndarray
        Generated surface GRFs
    """
    
    mu, sigma, tau = params
    rs = np.random.default_rng(seed)

    acorr, n = np.inf, 0
    while np.abs(np.abs(acorr) - corr) > tol:
        if alpha > 0:
            x, y = make_correlated_exgaussian(corr, size=902629,
                                      seed=rs.integers(MSEED), 
                                      mu=mu, sigma=sigma, tau=tau)
            # smooth correlated noise vectors + project to surface
            xs = create_surface_exgaussian(noise=x, alpha=alpha, normalize=normalize)
            ys = create_surface_exgaussian(noise=y, alpha=alpha, normalize=normalize)
        else:
            xs, ys = make_correlated_exgaussian(corr, size=20484,
                                        seed=rs.integers(MSEED),
                                        mu=mu, sigma=sigma, tau=tau)

        # remove medial wall to ensure data are still sufficiently correlated.
        # this is important for parcellations that will ignore the medial wall
        xs, ys = _mod_medial(xs, remove=True), _mod_medial(ys, remove=True)
        acorr = np.corrcoef(xs, ys)[0, 1]

        if debug:
            # n:>3 because dear lord i hope it doesn't take more than 999 tries
            print(f'{n:>3}: {acorr:>6.3f}')
            n += 1

    if acorr < 0:
        ys *= -1

    if normalize:
        xs, ys = sstats.zscore(xs), sstats.zscore(ys)

    return _mod_medial(xs, remove=False), _mod_medial(ys, remove=False)


"""
Eigenmode resampling on the cortex. Use fit.surface_fit() to determine how many
modes to use to capture the empirical variogram.
"""

from eigenstrapping import (
    SurfaceEigenstrapping, VolumetricEigenstrapping, compute_psd, utils,
    )
import matplotlib.pyplot as plt
import numpy as np
from netneurotools.stats import efficient_pearsonr
from brainsmash.mapgen import Base, Sampled
from brainsmash.mapgen.memmap import txt2memmap
from .geometry import geodesic_distmat
from .utils import parcellate_distmat, calc_parcellate
import os

eigen_args = ['surface', 'evals', 'emodes',
               'num_modes', 'save_surface',
               'seed', 'decomp_method', 'medial',
               'randomize', 'find_optimal', 'resample',
               'step', 'n_jobs', 'use_cholmod', 'permute',
               'distribution', 'shuffle', 'gen_rotations',
               'add_res', 'parcellation']

var_args = ['ns', 'pv', 'nh', 'knn', 'pv', 'nh', 'knn', 'n_jobs', 'seed']

def surface_fit(x, D=None, index=None, nsurrs=10, num_modes=100, return_data=False, 
                extra_diags=False, **params):
    """
    Evaluate variogram fits for :class: `eigenstrapping.SurfaceEigenstrapping` 
    to determine how many modes to decompose surface map with. Returns two plots:
        left - variogram fit
        middle - surrogate pairwise correlation with original map `x`
        right - histogram of surrogate residuals with original map `x`

    Parameters
    ----------
    x : (N,) np.ndarray
        Target brain map
    D : (N,N) np.ndarray or np.memmap
        Pairwise distance matrix between elements of `x`
    index : (N,N) np.ndarray or np.memmap
        See :method:`variogram.variogram`
    nsurr : int, default 20
        Number of simulated surrogate maps from which to compute variograms
    return_data : bool, default False
        if True, return the surrogates
    extra_diags : bool, default False
        if True, return extra diagnostics
            1. modal power spectra of original and surrogates
            2. Local Moran's I of original and surrogates
    params :
        Keyword arguments for :class: `eigenstrapping.SurfaceEigenstrapping` or
        :method: `variogram.variogram`. If eigenmodes and eigenvalues have been
        precomputed for the surface that `x` exists on, then pass them as `evals`
        and `emodes`. If they're not given, then a `surface` must be given.
        See :class: `eigenstrapping.Eigenstrapping` for more details.

    Returns
    -------
    if and only if return_data is True:
    emp_var : (M,) np.ndarray
        empirical smoothed variogram values
    u0 : (M,) np.ndarray
        distances at which variogram values were computed
    surr_var : (nsurr, M) np.ndarray
        surrogate maps' smoothed variogram values

    Notes
    -----
    If `return_data` is False, this function generates and shows a matplotlib
    plot instance illustrating the fit of the surrogates' variograms to the
    target map's variogram. If `return_data` is True, this function returns the
    data needed to generate such a plot (i.e., the variogram values and the
    corresponding distances).

    """
    # check kwargs
    eigen_params = dict()
    var_params = dict()
    for arg in params:
        if arg in eigen_args:
            eigen_params[arg] = params[arg]
        if arg in var_args:
            var_params[arg] = params[arg]
            
    # initialize
    eigen = SurfaceEigenstrapping(x, num_modes=num_modes, **eigen_params)
    
    # surrogates
    surrs = eigen(n=nsurrs)
    
    # plot variogram
    # Instantiate surrogate map generator
    print('Surrogates computed, computing stats...')
    if eigen_params['parcellation']:
        parcellation = eigen_params['parcellation']
        x = calc_parcellate(parcellation, data_input=x)
        surrs = calc_parcellate(parcellation, data_input=surrs).T
        generator = Base(x=x, D=D)
        
        emp_var, u0 = generator.compute_smooth_variogram(x, return_h=True)

        # Compute surrogate map variograms
        surr_var = np.empty((nsurrs, generator.nh))
        for i in range(nsurrs):
            surr_var[i] = generator.compute_smooth_variogram(surrs[i])
        
    else:
        if D or index is None:
            print('No surface distance matrix given, calculating - may take a while...')
            distmat = geodesic_distmat(eigen_params['surface'], n_jobs=eigen_params['n_jobs'])
            d = os.path.dirname(__file__)
            file = os.path.join(d, 'distmat.txt')
            np.savetxt(file, distmat)
            dict_dist = txt2memmap(file, output_dir=d)
            D = np.load(dict_dist['distmat'])
            index = np.load(dict_dist['index'])
        generator = Sampled(x=x, D=D, index=index)
    
        nsurrs_var = nsurrs
        if nsurrs > 50:
            nsurrs_var = 50
        # Compute target & surrogate map variograms
        surr_var = np.empty((nsurrs_var, generator.nh))
        emp_var_samples = np.empty((nsurrs_var, generator.nh))
        u0_samples = np.empty((nsurrs_var, generator.nh))
        for i in range(nsurrs_var):
            idx = generator.sample()  # Randomly sample a subset of brain areas
            v = generator.compute_variogram(generator.x, idx)
            u = generator.D[idx, :]
            umax = np.percentile(u, generator.pv)
            uidx = np.where(u < umax)
            emp_var_i, u0i = generator.smooth_variogram(
                u=u[uidx], v=v[uidx], return_h=True)
            emp_var_samples[i], u0_samples[i] = emp_var_i, u0i
            # Surrogate
            v_null = generator.compute_variogram(surrs[i], idx)
            surr_var[i] = generator.smooth_variogram(
                u=u[uidx], v=v_null[uidx], return_h=False)

        u0 = u0_samples.mean(axis=0)
        emp_var = emp_var_samples.mean(axis=0)
    
    #u0, emp_var, surr_var = variogram(x, surrs, D, index, **var_params)

    # Start with a 1x3 grid for plots
    fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(15, 5))
        
    # Variogram plot
    ax = axes[0]
    ax.scatter(u0, emp_var, s=20, facecolor='none', edgecolor='k',
               marker='o', lw=1, label='Empirical')
    mu = surr_var.mean(axis=0)
    sigma = surr_var.std(axis=0)
    ax.fill_between(u0, mu-sigma, mu+sigma, facecolor='#377eb8',
                    edgecolor='none', alpha=0.3)
    ax.plot(u0, mu, color='#377eb8', label='Surrogates', lw=1)
    leg = ax.legend(loc=0)
    leg.get_frame().set_linewidth(0.0)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.set_xlabel("Spatial separation\ndistance")
    ax.set_ylabel("Variance")
    ax.set_title("Variogram")
        
    # Pairwise correlation plot
    ax = axes[1]
    correlations = efficient_pearsonr(x, surrs.T, nan_policy='omit')[0]
    ax.hist(correlations, bins=np.linspace(-1, 1, num=nsurrs//5), color='#377eb8', alpha=0.7)
    ax.axvline(x=np.mean(correlations), color='r', linestyle='dashed', linewidth=3)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.set_title("Distribution of correlations")
    ax.set_xlabel("Correlation coefficient")
    ax.set_ylabel("Frequency")
        
    # Histogram of residuals plot
    ax = axes[2]
    residuals = [x - surr for surr in surrs]
    for res in residuals:
        ax.hist(res, bins=50, alpha=0.5, histtype='stepfilled')
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.set_title("Histogram of Residuals")
    ax.set_xlabel("Residual value")
    ax.set_ylabel("Frequency")
    
    # Ensure spacing between plots is nice
    plt.tight_layout()
    
    plt.show()
    
    if extra_diags:
        # compute modal power spectra
        original_power = eigen.psd
        surr_psds = [compute_psd(surr[eigen.medial_wall], eigen._emodes) for surr in surrs]
        # calculate moran
        original_moran = utils.calc_moran(D, x.reshape(1, -1), medmask=eigen_params['medial'] if 'medial'
                                                            in eigen_params else None)
        surr_moran = utils.calc_moran(D, surrs[:nsurrs//2], medmask=eigen_params['medial'] if 'medial'
                                                            in eigen_params else None)
        
        # now plot
        fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(10, 5))

        # Power Spectrum
        ax = axes[0]
        ax.semilogy(original_power, label='Original', color='k', linewidth=2, zorder=1000)
        for i, power in enumerate(surr_psds[:len(surr_psds)//2]):
            ax.semilogy(power, alpha=0.3, linestyle='dashed')
        ax.set_title('Power Spectrum')
        ax.set_xlabel('Mode')
        ax.set_ylabel('Normalized power')
        leg = ax.legend(loc=0)
        leg.get_frame().set_linewidth(0.0)
        
        # Moran's I
        ax = axes[1]
        labels = ['Original'] + [f'Surrogate {i+1}' for i in range(nsurrs//2)]
        
        # Only displaying every second label for clarity
        visible_labels = [label if idx % 2 == 0 else '' for idx, label in enumerate(labels)]
        
        ax.bar(range(len(labels)), np.hstack((original_moran, surr_moran)), color='c', alpha=0.7)
        ax.set_title("Moran's I")
        ax.set_ylabel("Moran's I Value")
        ax.set_xticks(range(len(labels)))
        ax.set_xticklabels(visible_labels, rotation=30, ha='right')
        
        plt.tight_layout()
        plt.show()
        
    if return_data:
        return surrs
    
def volumetric_fit(volume, x, nsurrs=10, num_modes=100, 
                   return_data=False, extra_diags=False, **params):
    """
    Evaluate variogram fits for :class: `eigenstrapping.VolumetricEigenstrapping` 
    to determine how many modes to decompose volumetric map with.

    Parameters
    ----------
    mask : str to mask file
        Target ROI volume with 1 inside region-of-interest, and 0 elsewhere
    x : str to data inside mask
        Target brain map
    D : (N,N) np.ndarray or np.memmap
        Pairwise distance matrix between elements of `x`
    index : (N,N) np.ndarray or np.memmap
        See :method:`variogram.variogram`
    nsurr : int, default 20
        Number of simulated surrogate maps from which to compute variograms
    return_data : bool, default False
        if True, return: 1, the smoothed variogram values for the target
        brain map; 2, the distances at which the smoothed variograms values
        were computed; and 3, the surrogate maps' smoothed variogram values
    params :
        Keyword arguments for :class: `eigenstrapping.VolumetricEigenstrapping` or
        :method: `variogram.variogram`. If eigenmodes and eigenvalues have been
        precomputed for the surface that `x` exists on, then pass them as `evals`
        and `emodes`. If they're not given, then a `surface` must be given.
        See :class: `eigenstrapping.VolumetricEigenstrapping` for more details.

    Returns
    -------
    if and only if return_data is True:
    surrs : (N,) np.ndarray
        Eigenstrapped surrogates in volume space

    """
    
    #check kwargs
    eigen_params = dict()
    var_params = dict()
    for arg in params:
        if arg in eigen_args:
            eigen_params[arg] = params[arg]
        if arg in var_args:
            var_params[arg] = params[arg]
            
    # initialize
    eigen = VolumetricEigenstrapping(volume, x, num_modes=num_modes, **eigen_params)
    
    # surrogates
    surrs = eigen(n=nsurrs)
    
    x, D, index = eigen.calculate_distance_matrix(return_data=True)
    
    # Instantiate surrogate map generator
    generator = Sampled(x=x, D=D, index=index, knn=900)

    if nsurrs > 50:
        nsurrs = 50
    # Compute target & surrogate map variograms
    surr_var = np.empty((nsurrs, generator.nh))
    emp_var_samples = np.empty((nsurrs, generator.nh))
    u0_samples = np.empty((nsurrs, generator.nh))
    for i in range(nsurrs):
        idx = generator.sample()  # Randomly sample a subset of brain areas
        v = generator.compute_variogram(generator.x, idx)
        u = generator.D[idx, :]
        umax = np.percentile(u, generator.pv)
        uidx = np.where(u < umax)
        emp_var_i, u0i = generator.smooth_variogram(
            u=u[uidx], v=v[uidx], return_h=True)
        emp_var_samples[i], u0_samples[i] = emp_var_i, u0i
        # Surrogate
        v_null = generator.compute_variogram(surrs[i], idx)
        surr_var[i] = generator.smooth_variogram(
            u=u[uidx], v=v_null[uidx], return_h=False)

    u0 = u0_samples.mean(axis=0)
    emp_var = emp_var_samples.mean(axis=0)

    # Start with a 1x3 grid for plots
    fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(15, 5))
        
    # Variogram plot
    ax = axes[0]
    ax.scatter(u0, emp_var, s=20, facecolor='none', edgecolor='k',
               marker='o', lw=1, label='Empirical')
    mu = surr_var.mean(axis=0)
    sigma = surr_var.std(axis=0)
    ax.fill_between(u0, mu-sigma, mu+sigma, facecolor='#377eb8',
                    edgecolor='none', alpha=0.3)
    ax.plot(u0, mu, color='#377eb8', label='Surrogates', lw=1)
    ax.legend(loc=0)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.set_xlabel("Spatial separation\ndistance")
    ax.set_ylabel("Variance")
    ax.set_title("Variogram")
        
    # Pairwise correlation plot
    ax = axes[1]
    correlations, _ = efficient_pearsonr(x, surrs.T, nan_policy='omit')
    ax.hist(correlations, bins=nsurrs//2, color='#377eb8', alpha=0.7)
    ax.axvline(x=np.mean(correlations), color='r', linestyle='dashed', linewidth=1)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.set_title("Pairwise Correlations")
    ax.set_xlabel("Correlation coefficient")
    ax.set_ylabel("Frequency")
        
    # Histogram of residuals plot
    ax = axes[2]
    residuals = [x - surr for surr in surrs]
    for res in residuals:
        ax.hist(res, bins=50, alpha=0.5, histtype='stepfilled')
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.set_title("Histogram of Residuals")
    ax.set_xlabel("Residual value")
    ax.set_ylabel("Frequency")
    
    # Ensure spacing between plots is nice
    plt.tight_layout()
    
    plt.show()
    
    if extra_diags:
        # compute modal power spectra
        original_power = eigen.psd
        surr_psds = [compute_psd(surr, eigen._emodes) for surr in surrs]
        # calculate moran
        original_moran = utils.calc_moran(D, x.reshape(1, -1))
        surr_moran = utils.calc_moran(D, surrs[:nsurrs//2])
        
        # now plot
        fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(10, 5))

        # Power Spectrum
        ax = axes[0]
        ax.semilogy(original_power, label='Original', color='k', linewidth=2, zorder=1000)
        for i, power in enumerate(surr_psds[:len(surr_psds)//2]):
            ax.semilogy(power, alpha=0.3, linestyle='dashed')
        ax.set_title('Power Spectrum')
        ax.set_xlabel('Mode')
        ax.set_ylabel('Normalized power')
        leg = ax.legend(loc=0)
        leg.get_frame().set_linewidth(0.0)
        
        # Moran's I
        ax = axes[1]
        labels = ['Original'] + [f'Surrogate {i+1}' for i in range(nsurrs//2)]
        
        # Only displaying every second label for clarity
        visible_labels = [label if idx % 2 == 0 else '' for idx, label in enumerate(labels)]
        
        ax.bar(range(len(labels)), np.hstack((original_moran, surr_moran)), color='c', alpha=0.7)
        ax.set_title("Moran's I")
        ax.set_ylabel("Moran's I Value")
        ax.set_xticks(range(len(labels)))
        ax.set_xticklabels(visible_labels, rotation=30, ha='right')
        
        plt.tight_layout()
        plt.show()
    
    if return_data:
        return surrs


import numpy as np
from . import spatial, parutils, stats
from .geometry import read_annot
from joblib import Parallel, delayed, dump, load
from pathlib import Path
from netneurotools import datasets as nndata
import os.path as op
from nilearn import image
from scipy.sparse import csr_matrix


"""
Eigenmode helper functions (C) Systems Neuroscience Newcastle &
Nikitas C. Koussis 2023
"""

DROP = [  # regions that should always be dropped from analyses
    'lh_unknown', 'rh_unknown',
    'lh_corpuscallosum', 'rh_corpuscallosum',
    'lh_Background+FreeSurfer_Defined_Medial_Wall',
    'rh_Background+FreeSurfer_Defined_Medial_Wall',
]

def is_string_like(obj):
    """ Check whether `obj` behaves like a string. """
    try:
        obj + ''
    except (TypeError, ValueError):
        return False
    return True

def get_schaefer(vers='fsaverage5', data_dir=None, networks='7', scale=400):
    """
    Returns CSchaefer 2018 atlases as dictionary

    Parameters
    ----------
    vers : str, optional
        Which version of the atlases to get. Default: 'fsaverage5'
    data_dir : str or os.PathLike, optional
        Data directory where downloaded atlases should be stored. If not
        specified will default to $NNT_DATA or ~/nnt-data
    networks : {'7', '17'}, optional
        Which networks to get for Schaefer 2018 atlas. Default: '7'

    Returns
    -------
    list of filenames
        Filenames where left [0] and right [1] hemisphere labels are kept
    """
    schaefer = nndata.fetch_schaefer2018(vers, data_dir=data_dir)
    schaefer = {k: schaefer.get(k) for k in schaefer.keys()
                if f'{scale}Parcels7Networks' in k}

    return schaefer[f'{scale}Parcels7Networks']

def parcify(annotation, drop=True):
    """
    Returns an array of parcel labels for every vertex in the surface
    of `annotation`

    Parameters
    ----------
    annotation : str
        FreeSurfer annotation file
    drop : bool, optional
        Whether to drop the medial wall labels. Default True.

    Returns
    -------
    parc : np.ndarray of shape (N,)
        Parcel labels for every vertex

    """
    
    labels, ctab, names = read_annot(annotation)
    parc = labels
    
    if drop:
        drop = np.intersect1d(names, DROP)
        parc = np.delete(labels, [names.index(f) for f in drop])
        
    return parc


def _get_eigengroups(eigs, suppress_message=False):
    """
    Helper function to find eigengroups
    """
    if suppress_message is False:
        print("IMPORTANT: EIGENMODES MUST BE TRUNCATED AT FIRST NON-ZERO MODE FOR THIS FUNCTION TO WORK")
    lam = eigs.shape[1] # number of eigenmodes, accounting for discarded non-zero mode
    l = np.floor((lam-1)/2).astype(int)    
    # Handle cases where lam is from 4 to 7
    if lam == 3:
        return [np.arange(0, 3)]
    elif 4 <= lam < 8:
        return [np.arange(0, 3), np.arange(3, lam)]
    elif lam < 3:
        raise ValueError('Number of modes to resample cannot be less than 3')
    
    groups = []#[np.array([0])]
    ii = 0
    i = 0
    for g in range(1, l):
        ii += 2*g+1
        if ii >= lam:
            groups.append(np.arange(i,lam))
            return groups
        groups.append(np.arange(i,ii))
        i = ii
    

def reconstruct_data(coeffs, eigenmodes):
    """
    Reconstruct a dataset of `n_vertices` given a set of eigenmodes and coeffs
    conditioned on data using ordinary least squares (OLS)

    Parameters
    ----------
    coeffs : np.ndarray of shape (M,)
        Coefficients output from fitting OLS
    eigenmodes : np.ndarray of shape (n_vertices, M)
        Eigenmodes of `n_vertices` by number of eigenvalues M
    pv : float (default 100), rounds up to nearest index
        Percentage of `eigenmodes` to use for reconstruction (default 100%)

    Returns
    -------
    new_data : np.ndarray of (n_vertices,)
        Reconstructed data

    """
    
    coeffs = coeffs.reshape(-1, 1)
    
    new_data = eigenmodes @ coeffs
    
    return new_data.squeeze()
    
def eigen_decomposition(data, eigenmodes, method='matrix'):
    """
    Decompose data using eigenmodes and calculate the coefficient of 
    contribution of each vector
    
    Parameters:
    -----------
    data : np.ndarray of shape (n_vertices, 3)
        N = number of vertices, P = columns of independent data
    eigenmodes : np.ndarray of shape (n_vertices, M)
        N = number of vertices, M = number of eigenmodes
    method : string
        method of calculation of coefficients: 'matrix', 'matrix_separate', 
        'regression'
    
    Returns:
    -------
    coeffs : numpy array of shape (N, 3)
     coefficient values
    
    """
    
    if data.ndim > 1:
        N, P = data.shape
    else:
        P = 1
    
    _, M = eigenmodes.shape
    
    if method == 'matrix':
        #print("Using matrix decomposition to reconstruct data")
        coeffs = np.linalg.solve((eigenmodes.T @ eigenmodes), (eigenmodes.T @ data))
    elif method == 'matrix_separate':
        coeffs = np.zeros((M, P))
        if P > 1:
            for p in range(P):
                coeffs[:, p] = np.linalg.solve((eigenmodes.T @ eigenmodes), (eigenmodes.T @ data[:, p]))
    elif method == 'regression':
        #print("Using regression method to reconstruct data")
        coeffs = np.zeros((M, P))
        if P > 1:
            for p in range(P):
                coeffs[:, p] = np.linalg.lstsq(eigenmodes, data[:, p], rcond=None)[0]
        else:
            coeffs = np.linalg.lstsq(eigenmodes, data, rcond=None)[0]
            
    else:
        raise ValueError("Accepted methods for decomposition are 'matrix', and 'regression'")
                
    return coeffs

def calc_surface_connectivity(data, parc):
    """
    Calculate surface-based connectivity matrix.    

    Parameters
    ----------
    data : ndarray of shape (N,)
        Surface data to calculate connectivity matrix
    parc : ndarray of shape (N,)
        Surface labels for each parcel that corresponds to an atlas
    
    Returns
    -------
    surface_connectivity : np.ndarray of shape (num_parcels, num_parcels)
        Surface connectivity matrix

    """

    parc_data = calc_parcellate(parc, data)
    matrix = stats.pairwise_r(parc_data)

    return matrix
    
    
def compute_axes_ellipsoid(eigenvalues):
    """
    Compute the axes of an ellipsoid given the eigenmodes.
    """    
    return np.sqrt(eigenvalues)
    

def transform_to_spheroid(eigenvalues, eigenmodes):
    """
    Transform the eigenmodes to a spheroid space
    """
    ellipsoid_axes = compute_axes_ellipsoid(eigenvalues)
    #ellipsoid_axes = ellipsoid_axes.reshape(-1, 1)
    
    spheroid_eigenmodes = np.divide(eigenmodes, np.sqrt(eigenvalues))
    
    return spheroid_eigenmodes
    
    
def transform_to_ellipsoid(eigenvalues, eigenmodes):
    """
    Transform the eigenmodes in spheroid space back to ellipsoid by stretching
    """
    
    ellipsoid_axes = compute_axes_ellipsoid(eigenvalues)
    
    ellipsoid_eigenmodes = np.multiply(eigenmodes, np.sqrt(eigenvalues))
    
    return ellipsoid_eigenmodes


def resample_spheroid(spheroid_eigenmodes, angle):
    """
    Resample the N-D hypersphere generated by the N orthogonal unit modes

    """    
    # initialize the new points p
    p = spheroid_eigenmodes * np.cos(angle)
    
    new_modes = p
    
    return new_modes

def calc_parcellate(parc, data_input):
    """
    Calculate parcellated data given (N,) parcellation for every vertex of shape 
    (N,)

    Parameters
    ----------
    parc : str or np.ndarray (N,)
        Parcellation (np.ndarray of n_labels)
    data_input : np.ndarray (N, M)
        Data to parcellate

    Returns
    -------
    data_parcellated : (n_labels, M)
        Parcellated data

    """

    if is_string_like(parc):
        _, ext = op.splitext(parc)
        if ext == '.annot':
            parc = parcify(parc)
        elif ext == '.txt':
            parc = np.loadtxt(parc)
        elif ext == '.gii':
            parc = image.load_img(parc).agg_data()
        else:
            raise ValueError('Cannot read input parcellation')
            
    num_vertices = parc.shape[0]
    parcels = np.unique(parc[parc > 0])
    num_parcels = len(parcels)
    
    if data_input.ndim == 1:
        data_input = data_input.reshape(-1,1)

    if data_input.shape[0] != num_vertices:
        data_input = data_input.T

    data_parcellated = np.zeros((num_parcels, data_input.shape[1]))

    for parcel_ind in range(num_parcels):
        parcel_interest = parcels[parcel_ind]

        ind_parcel = np.where(parc == parcel_interest)[0]

        data_parcellated[parcel_ind] = np.nanmean(data_input[ind_parcel])

    return data_parcellated.squeeze()

def parcellate_distmat(D, parc):
    """
    Parcellate a dense distance matrix

    Parameters
    ----------
    D : np.ndarray of shape (N, N)
        Distance matrix of surface
    parc : np.ndarray of int of shape (N,)
        Vector of labels (ints) for each vertex on surface with `num_parcels`
        number of labels

    Returns
    -------
    D : ndarray of shape (num_parcels, num_parcels)
        Parcellated distance matrix

    """
    
    D = np.row_stack([
        calc_parcellate(parc, D[row]) for row in range(D.shape[0])
    ])
    D[np.diag_indices_from(D)] = 0
    D = D[1:, 1:] + D[1:, 1:].T
    
    return D

def normalize_data(data):
    data_normalized = np.subtract(data, np.nanmean(data, axis=0))
    data_normalized = np.divide(data_normalized, np.nanstd(data_normalized, axis=0))
    
    return data_normalized

def calc_moran(dist, nulls, medmask, n_jobs=1):
    """
    Calculates Moran's I for every column of `nulls`

    Parameters
    ----------
    dist : (N, N) array_like
        Hemispheric distance matrix
    nulls : (N, P) array_like
        Null brain maps for which to compute Moran's I
    medmask : (N,) array_like
        Medial wall mask array, where medial wall values are 0
    n_jobs : int, optional
        Number of parallel workers to use for calculating Moran's I. Default: 1

    Returns
    -------
    moran : (P,) np.ndarray
        Moran's I for `P` null maps
    """

    # calculate moran's I, masking out NaN values for each null (i.e., the
    # rotated medial wall)
    fn = dump(dist, spatial.make_tmpname('.mmap', prefix='/tmp/'))[0]
    dist = load(fn, mmap_mode='r')
    
    #do some pre-calculation on our distance matrix to reduce computation time
    with np.errstate(divide='ignore', invalid='ignore'):
        dist = 1 / dist
        np.fill_diagonal(dist, 0)
        dist /= dist.sum(axis=-1, keepdims=True)
        
    if medmask is None:
        medmask = np.isnan(nulls[:, 0])
        
    moran = np.array(
        Parallel(n_jobs=n_jobs)(
            delayed(_moran)(dist, nulls[:, n], medmask)
            for n in parutils.trange(nulls.shape[-1], desc="Running Moran's I")
        )
    )

    Path(fn).unlink()
    return moran

def _moran(dist, sim, medmask):
    mask = np.logical_and(medmask, np.logical_not(np.isnan(sim)))
    return spatial.morans_i(dist[np.ix_(mask, mask)], sim[mask],
                            normalize=False, invert_dist=False)



#from nipype.interfaces.freesurfer import MRIMarchingCubes
from lapy import TriaMesh, TriaIO, Solver
from lapy.DiffGeo import tria_compute_gradient, tria_compute_divergence
import warnings
from collections import OrderedDict
import scipy.optimize as optimize
import subprocess
import os
import nibabel as nib
#from nilearn import image
import numpy as np
#from neuromaps.datasets.atlases import fetch_mni152
#from ants import image_read, registration, apply_transforms
from scipy.spatial import Delaunay, KDTree
from brainspace import mesh
from joblib import Parallel, delayed
from scipy.sparse.linalg import splu

try:
    from sksparse.cholmod import cholesky
except:
    print("Scikit-sparse libraries not found, using LU decomposition for eigenmodes (slower)")


"""
Helper utilities for geometry and registration

    - Runs mri_mc from Freesurfer to create 2d surface
    - Projects 2d surface using gmsh
    - Writes out geometry in tetrahedral format, triangular format, or in Freesurfer binary
    - Writes out label files
    - Registers from native space to MNI152 and fsaverage space

read_geometry() was taken from nibabel.freesurfer package (https://github.com/nipy/nibabel/blob/master/nibabel/freesurfer/io.py).
get_tkrvox2ras(), make_tetra_file(), normalize_vtk(), and calc_volume() taken
from <https://github.com/BMHLab/BrainEigenmodes>, authors J. Pang and K. Aquino.

"""

_ANNOT_DT = ">i4"
"""Data type for Freesurfer `.annot` files.

Used by :func:`read_annot` and :func:`write_annot`.  All data (apart from
strings) in an `.annot` file is stored as big-endian int32.
"""

# mni152_2mm = np.asarray(
#                  [[  2.,    0.,   0.,    -90,],
#                   [ -0.,    2.,   0.,   -126,],
#                   [ -0.,    0.,   2.,    -72,],
#                   [  0.,    0.,   0.,      1.]])

# mni152_1mm = np.asarray(
#                 [[  -1.,    0.,    0.,   90.],
#                  [   0.,    1.,    0., -126.],
#                  [   0.,    0.,    1.,  -72.],
#                  [   0.,    0.,    0.,    1.]])

# def _check_mni(in_file):
#     """
#     Checks if input image is in MNI152 space
#     """
    
#     img = image.load_img(in_file)
    
#     if img.affine != mni152_2mm:
#         if img.affine != mni152_1mm:
#             return False
        
#     else:
#         return True
    
# def native_to_mni152(in_file, nonlinear=True):
#     """
#     Linear or nonlinear registration of native volumetric image to MNI152 space
#     Uses ANTsPy
#     """
    
#     img = image_read(in_file)
    
#     # get template image
#     mni_file = fetch_mni152(density='1mm').get('2009cAsym_T1w')
#     mni = image_read(mni_file)

#     if nonlinear is True:
#         transform_type='SyN'
        
#     else:
#         transform_type='Affine'
        
#     # do transform
#     fixed_image = mni
#     moving_image = img
    
#     mytx = registration(fixed=fixed_image, moving=moving_image, type_of_transform=transform_type)
    
#     warped_moving_image = apply_transforms(fixed=fixed_image, moving=moving_image,
#                                            transformlist=mytx['fwdtransforms'])
    
#     # rebuild as nib.Nifti1Image
#     transformed_image = warped_moving_image.to_nibabel()
    
#     return transformed_image

def gaussian(x, amplitude, mean, stddev):
    return amplitude * np.exp(- ((x - mean) ** 2) / (2 * (stddev ** 2)))

def estimate_fwhm(image):
    # Find the maximum value and its position in the image
    max_val = np.max(image)
    max_pos = np.argmax(image)
    max_row, max_col = np.unravel_index(max_pos, image.shape)
    
    # Create a 1d profile along the maximum row and column
    row_profile = image[max_row, :]
    col_profile = image[:, max_col]
    
    options = {'maxfev' : np.prod(image.shape)}
    
    # Fit a gaussian curve to the row profile
    row_x = np.arange(len(row_profile))
    row_params, _ = optimize.curve_fit(gaussian, row_x, row_profile, p0=[max_val, max_col, 1], **options)
    
    # Fit a gaussian curve to the column profile
    col_x = np.arange(len(col_profile))
    col_params, _ = optimize.curve_fit(gaussian, col_x, col_profile, p0=[max_val, max_row, 1], **options)
    
    # Calculate the FWHM as 2.355 times the std dev of the Gaussian
    row_fwhm = 2.355 * row_params[2]
    col_fwhm = 2.355 * col_params[2]
    
    # take the average
    fwhm = (row_fwhm + col_fwhm) / 2
    
    return fwhm
    
def resel_count(image, fwhm):
    # returns resel count of image given FWHM
    return np.prod(image.shape / fwhm)
        
def get_tkrvox2ras(voldim, voxres):
    """Generate transformation matrix to switch between tetrahedral and volume space.

    Parameters
    ----------
    voldim : array (1x3)
        Dimension of the volume (number of voxels in each of the 3 dimensions)
    voxres : array (!x3)
        Voxel resolution (resolution in each of the 3 dimensions)

    Returns
    ------
    T : array (4x4)
        Transformation matrix
    """

    T = np.zeros([4,4]);
    T[3,3] = 1;

    T[0,0] = -voxres[0];
    T[0,3] = voxres[0]*voldim[0]/2;

    T[1,2] = voxres[2];
    T[1,3] = -voxres[2]*voldim[2]/2;


    T[2,1] = -voxres[1];
    T[2,3] = voxres[1]*voldim[1]/2;

    return T

def make_tetra_file(nifti_input_filename):
    """Generate tetrahedral version of the ROI in the nifti file.

    Parameters
    ----------
    nifti_input_filename : str
        Filename of input volume where the relevant ROI have voxel values = 1

    Returns
    ------
    tetra_file : str
        Filename of output tetrahedral vtk file
    """

    nifti_input_file_head, nifti_input_file_tail = os.path.split(nifti_input_filename)
    nifti_input_file_main, nifti_input_file_ext = os.path.splitext(nifti_input_file_tail)

    os.system('mri_mc ' + nifti_input_filename + ' 1 ' + nifti_input_file_head + '/rh.tmp_surface.vtk')
    os.system('mv -f ' + nifti_input_file_head + '/rh.tmp_surface.vtk ' + nifti_input_filename + '.vtk')

    geo_file = nifti_input_filename + '.geo'
    tria_file = nifti_input_filename + '.vtk'
    tetra_file = nifti_input_filename + '.tetra.vtk'

    file = tria_file.rsplit('/')
    inputGeo = file[len(file)-1]
    
    with open(geo_file, 'w') as writer:
        writer.write('Mesh.Algorithm3D=4;\n')
        writer.write('Mesh.Optimize=1;\n')
        writer.write('Mesh.OptimizeNetgen=1;\n')
        writer.write('Merge "'+inputGeo+'";\n')
        writer.write('Surface Loop(1) = {1};\n')
        writer.write('Volume(1) = {1};\n')
        writer.write('Physical Volume(1) = {1};\n')

    cmd = 'gmsh -3 -o ' + tetra_file + ' ' + geo_file
    output = subprocess.check_output(cmd,shell="True")
    output = output.splitlines()

    cmd = "sed 's/double/float/g;s/UNSTRUCTURED_GRID/POLYDATA/g;s/CELLS/POLYGONS/g;/CELL_TYPES/,$d' " + tetra_file + " > " + tetra_file + "'_fixed'"
    os.system(cmd)
    os.system('mv -f ' + tetra_file + '_fixed ' + tetra_file)
    
    return tetra_file

def make_tria_file(nifti_input_filename):
    """
    Generate triangular mesh of the ROI in the nifti file.

    Parameters
    ----------
    nifti_input_filename : str
        Filename of input volume where the relevant ROI have voxel values = 1

    Returns
    -------
    tria_file : str
        Filename of output triangular vtk file

    """
    nifti_input_file_head, nifti_input_file_tail = os.path.split(nifti_input_filename)
    nifti_input_file_main, nifti_input_file_ext = os.path.splitext(nifti_input_file_tail)

    os.system('mri_mc ' + nifti_input_filename + ' 1 ' + nifti_input_file_head + '/rh.tmp_surface.vtk')
    os.system('mv -f ' + nifti_input_file_head + '/rh.tmp_surface.vtk ' + nifti_input_filename + '.vtk')

    geo_file = nifti_input_filename + '.geo'
    tria_file = nifti_input_filename + '.vtk'

    file = tria_file.rsplit('/')
    inputGeo = file[len(file)-1]
    
    with open(geo_file, 'w') as writer:
        writer.write('Mesh.Algorithm3D=4;\n')
        writer.write('Mesh.Optimize=1;\n')
        writer.write('Mesh.OptimizeNetgen=1;\n')
        writer.write('Merge "'+inputGeo+'";\n')
        writer.write('Surface Loop(1) = {1};\n')
        writer.write('Volume(1) = {1};\n')
        writer.write('Physical Volume(1) = {1};\n')

    cmd = 'gmsh -2 -o ' + tria_file + ' ' + geo_file
    output = subprocess.check_output(cmd,shell="True")
    output = output.splitlines()

    cmd = "sed 's/double/float/g;s/UNSTRUCTURED_GRID/POLYDATA/g;s/CELLS/POLYGONS/g;/CELL_TYPES/,$d' " + tria_file + " > " + tria_file + "'_fixed'"
    os.system(cmd)
    os.system('mv -f ' + tria_file + '_fixed ' + tria_file)
    
    return tria_file

def create_temp_surface(surface_input, surface_output_filename):
    """Write surface to a new vtk file.

    Parameters
    ----------
    surface_input : brainspace compatible object
        Loaded vtk object corresponding to a surface triangular mesh
    surface_output_filename : str
        Filename of surface to be saved
    """

    f = open(surface_output_filename, 'w')
    f.write('# vtk DataFile Version 2.0\n')
    f.write(surface_output_filename + '\n')
    f.write('ASCII\n')
    f.write('DATASET POLYDATA\n')
    f.write('POINTS ' + str(np.shape(surface_input.Points)[0]) + ' float\n')
    for i in range(np.shape(surface_input.Points)[0]):
        f.write(' '.join(map(str, np.array(surface_input.Points[i, :]))))
        f.write('\n')
    f.write('\n')
    f.write('POLYGONS ' + str(np.shape(surface_input.polys2D)[0]) + ' ' + str(4* np.shape(surface_input.polys2D)[0]) + '\n')
    for i in range(np.shape(surface_input.polys2D)[0]):
        f.write(' '.join(map(str, np.append(3, np.array(surface_input.polys2D[i, :])))))
        f.write('\n')
    f.close()

def get_indices(surface_original, surface_new):
    """Extract indices of vertices of the two surfaces that match.

    Parameters
    ----------
    surface_original : brainspace compatible object
        Loaded vtk object corresponding to a surface triangular mesh
    surface_new : brainspace compatible object
        Loaded vtk object corresponding to a surface triangular mesh

    Returns
    ------
    indices : array
        indices of vertices
    """

    indices = np.zeros([np.shape(surface_new.Points)[0],1])
    for i in range(np.shape(surface_new.Points)[0]):
        indices[i] = np.where(np.all(np.equal(surface_new.Points[i,:],surface_original.Points), axis=1))[0][0]
    indices = indices.astype(int)
    
    return indices

def calc_eig(tria, num_modes, use_cholmod=False):
    """Calculate the eigenvalues and eigenmodes of a surface.

    Parameters
    ----------
    tria : lapy compatible object
        Loaded vtk object corresponding to a surface triangular mesh
    num_modes : int
        Number of eigenmodes to be calculated

    Returns
    ------
    evals : array (num_modes x 1)
        Eigenvalues
    emodes : array (number of surface points x num_modes)
        Eigenmodes
    """
    
    fem = Solver(tria, use_cholmod=use_cholmod)
    evals, emodes = fem.eigs(k=num_modes)
    
    return evals, emodes
    
def calc_surface_eigenmodes(surface_input_filename, mask_input, save_cut=False, num_modes=200, use_cholmod=False, remove_zero=True):
    """Main function to calculate the eigenmodes of a cortical surface with 
    application of a mask (e.g., to remove the medial wall).

    Parameters
    ----------
    surface_input_filename : str
        Filename of input surface with ``n_vertices``
    mask_input : str or 1D ndarray
        Filename of mask to be applied on the surface (e.g., cortex without 
        medial wall, values = 1 for mask and 0 elsewhere)
    save_cut : boolean 
        Boolean to decide if the new surface with mask applied will be saved to 
        a new surface file
    num_modes : int
        Number of eigenmodes to be calculated      
    use_cholmod : bool, optional
        Spec to use cholesky libraries of ``scikit-sparse``. Requires installation
        of ``scikit-sparse`` and ``libsuitesparse-dev``.
        
    Returns
    -------
    np.ndarray of shape=(num_modes,)
        Eigenvalues computed on masked surface.
    np.ndarray of shape=(n_vertices, num_modes)
        Eigenmodes corresponding to eigenvalues computed on masked surface
        padded with zeros to match size of original surface (``n_vertices``)
    """

    # load surface (as a brainspace object)
    surface_orig = mesh.mesh_io.read_surface(surface_input_filename)
    
    # load mask
    # can be any ROI (even whole cortex)
    if isinstance(mask_input, np.ndarray):
        mask = mask_input
    else:
        mask_input_file_main, mask_input_file_ext = os.path.splitext(mask_input)
        if mask_input_file_ext == '.txt':
            mask = np.loadtxt(mask_input)
        elif mask_input_file_ext == '.gii':
            mask = nib.load(mask_input).darrays[0].data
        else:
            mask = mask_input
    
    # create temporary suface based on mask
    surface_cut = mesh.mesh_operations.mask_points(surface_orig, mask)

    if save_cut is True:
        # old method: save vtk of surface_cut and open via lapy TriaIO 
        # The writing phase of this process is very slow especially for large surfaces
        temp_cut_filename='temp_cut.gii'
        temp_cut = nib.GiftiImage()
        temp_cut.add_gifti_data_array(nib.gifti.gifti.GiftiDataArray(surface_cut.GetPoints().astype('float32')))
        temp_cut.add_gifti_data_array(nib.gifti.gifti.GiftiDataArray(mesh.mesh_elements.get_cells(surface_cut).astype('int32')))
        nib.save(temp_cut, temp_cut_filename)
        # load surface (as a lapy object)
        vertices, faces = temp_cut.agg_data()
        tria = TriaMesh(vertices, faces)
    else:
        # new method: replace v and t of surface_orig with v and t of surface_cut
        # faster version without the need to write the vtk file
        # load surface (as a lapy object)
        tria = TriaIO.import_vtk(surface_input_filename)
        tria.v = surface_cut.Points
        tria.t = np.reshape(surface_cut.Polygons, [surface_cut.n_cells, 4])[:,1:4]

    # calculate eigenvalues and eigenmodes
    evals, emodes = calc_eig(tria, num_modes, use_cholmod)
    
    if remove_zero:
        # remove zeroth eigenmode
        evals = evals[1:]
        emodes = emodes[:, 1:]
    
    # get indices of vertices of surface_orig that match surface_cut
    indices = get_indices(surface_orig, surface_cut)
    
    # reshape emodes to match vertices of original surface
    emodes_reshaped = np.zeros([surface_orig.n_points,np.shape(emodes)[1]])
    for mode in range(np.shape(emodes)[1]):
        emodes_reshaped[indices,mode] = np.expand_dims(emodes[:,mode], axis=1)
        
    return evals, emodes_reshaped

def compute_normals(coords, faces):
    """
    Compute the normal for each vertex in the mesh.
    """

    normals = np.zeros(coords.shape, dtype=coords.dtype)
    for i in range(faces.shape[0]):
        fn = np.cross(coords[faces[i, 1]] - coords[faces[i, 0]], coords[faces[i, 2]] - coords[faces[i, 0]])
        normals[faces[i, 0]] += fn
        normals[faces[i, 1]] += fn
        normals[faces[i, 2]] += fn

    norms = np.sqrt((normals ** 2).sum(axis=1))
    normals /= norms[:, np.newaxis]

    return normals

def inflate_mesh(vertices, faces, iterations=10, inflation_rate=0.02):
    """
    Inflate the mesh using basic inflation algorithm.
    """
    new_vertices = vertices
    for _ in range(iterations):

        normals = compute_normals(vertices, faces)

        new_vertices += normals * inflation_rate

    return new_vertices   

def nearest_neighbor(P, X, radius=None):
    """
    Find the one-nearest neighbors of vertices in points `P` on another 
    surface `X` using Delaunay triangulation and KDTree query.

    Parameters
    ----------
    P : np.ndarray of shape (N,3)
        Points to search for within the coordinate set of `X`. `P` can
        be a single point
    X : np.ndarray of shape (M,3)
        Vertices of the surface to search within
    radius : float
        Radius to search for nearest neighbors within

    Returns
    -------
    nearest_indexes : int
        Indexes of one-nearest neighbors of vertices in `P`. Note that
        if two vertices in `X` are the same distance away from a point in `P`,
        function returns only the first one.

    """
    
    # Create Delaunay triangulation for first surface
    tri = Delaunay(X)
    
    # Create tree of vertices to query on
    kdtree = KDTree(X)

    indices = np.empty(P.shape[0], dtype=int)
    for i, p in enumerate(P):
        simplex_index = tri.find_simplex(p)
        if simplex_index == -1 or (radius is not None and not _is_point_within_radius(p, X[tri.simplices[simplex_index]], radius)):
            _, nearest_neighbor_index = kdtree.query(p)
        else:
            simplex_vertices = X[tri.simplices[simplex_index]]
            dist = np.linalg.norm(simplex_vertices - p, axis=1)
            if radius is not None:
                valid_indices = np.where(dist <= radius)[0]
                if valid_indices.size == 0:
                    _, nearest_neighbor_index = kdtree.query(p)
                else:
                    nearest_neighbor_index = tri.simplices[simplex_index][valid_indices[np.argmin(dist[valid_indices])]]
            else:
                nearest_neighbor_index = tri.simplices[simplex_index][np.argmin(dist)]
        indices[i] = nearest_neighbor_index

    return indices


def _is_point_within_radius(p, vertices, radius):
    """
    Check if a point is within a given radius of any vertex in a set of vertices.
    """
    return np.any(np.linalg.norm(vertices - p, axis=1) <= radius)


def calc_volume(nifti_input_filename):
    """Calculate the physical volume of the ROI in the nifti file.

    Parameters
    ----------
    nifti_input_filename : str
        Filename of input volume where the relevant ROI have voxel values = 1

    Returns
    ------
    ROI_number : int
        Total number of non-zero voxels
    ROI_volume : float
        Total volume of non-zero voxels in physical dimensions   
    """

    # Load data
    ROI_data = nib.load(nifti_input_filename)
    roi_data = ROI_data.get_fdata()

    # Get voxel dimensions in mm
    voxel_dims = (ROI_data.header["pixdim"])[1:4]
    voxel_vol = np.prod(voxel_dims)

    # Compute volume
    ROI_number = np.count_nonzero(roi_data)
    ROI_volume = ROI_number * voxel_vol

    # print("Number of non-zero voxels = {}".format(ROI_number))
    # print("Volume of non-zero voxels = {} mm^3".format(ROI_volume))

    return ROI_number, ROI_volume

def normalize_vtk(tet, nifti_input_filename, normalization_type='none', normalization_factor=1):
    """Normalize tetrahedral surface.

    Parameters
    ----------
    tet : lapy compatible object
        Loaded vtk object corresponding to a surface tetrahedral mesh
    nifti_input_filename : str
        Filename of input volume where the relevant ROI have voxel values = 1
    normalization_type : str (default: 'none')
        Type of normalization
        number - normalization with respect to the total number of non-zero voxels
        volume - normalization with respect to the total volume of non-zero voxels in physical dimensions   
        constant - normalization with respect to a chosen constant
        others - no normalization
    normalization_factor : float (default: 1)
        Factor to be used in a constant normalization     

    Returns
    ------
    tet_norm : lapy compatible object
        Loaded vtk object corresponding to the normalized surface tetrahedral mesh
    """

    nifti_input_file_head, nifti_input_file_tail = os.path.split(nifti_input_filename)
    nifti_input_file_main, nifti_input_file_ext = os.path.splitext(nifti_input_file_tail)

    ROI_number, ROI_volume = calc_volume(nifti_input_filename)

    # normalization process
    tet_norm = tet
    if normalization_type == 'number':
        tet_norm.v = tet.v/(ROI_number**(1/3))
    elif normalization_type == 'volume':
        tet_norm.v = tet.v/(ROI_volume**(1/3))
    elif normalization_type == 'constant':
        tet_norm.v = tet.v/(normalization_factor**(1/3))
    else:
        pass

    # writing normalized surface to a vtk file
    if normalization_type == 'number' or normalization_type == 'volume' or normalization_type == 'constant':
        surface_output_filename = nifti_input_filename + '_norm=' + normalization_type + '.tetra.vtk'

        f = open(surface_output_filename, 'w')
        f.write('# vtk DataFile Version 2.0\n')
        f.write(nifti_input_file_tail + '\n')
        f.write('ASCII\n')
        f.write('DATASET POLYDATA\n')
        f.write('POINTS ' + str(np.shape(tet.v)[0]) + ' float\n')
        for i in range(np.shape(tet.v)[0]):
            f.write(' '.join(map(str, tet_norm.v[i, :])))
            f.write('\n')
        f.write('\n')
        f.write('POLYGONS ' + str(np.shape(tet.t)[0]) + ' ' + str(5 * np.shape(tet.t)[0]) + '\n')
        for i in range(np.shape(tet.t)[0]):
            f.write(' '.join(map(str, np.append(4, tet.t[i, :]))))
            f.write('\n')
        f.close()

    return tet_norm

def read_annot(filepath, orig_ids=False):
    """Read in a Freesurfer annotation from a ``.annot`` file.

    An ``.annot`` file contains a sequence of vertices with a label (also known
    as an "annotation value") associated with each vertex, and then a sequence
    of colors corresponding to each label.

    Annotation file format versions 1 and 2 are supported, corresponding to
    the "old-style" and "new-style" color table layout.

    Note that the output color table ``ctab`` is in RGBT form, where T
    (transparency) is 255 - alpha.

    See:
     * https://surfer.nmr.mgh.harvard.edu/fswiki/LabelsClutsAnnotationFiles#Annotation
     * https://github.com/freesurfer/freesurfer/blob/dev/matlab/read_annotation.m
     * https://github.com/freesurfer/freesurfer/blob/8b88b34/utils/colortab.c

    Parameters
    ----------
    filepath : str
        Path to annotation file.
    orig_ids : bool
        Whether to return the vertex ids as stored in the annotation
        file or the positional colortable ids. With orig_ids=False
        vertices with no id have an id set to -1.

    Returns
    -------
    labels : ndarray, shape (n_vertices,)
        Annotation id at each vertex. If a vertex does not belong
        to any label and orig_ids=False, its id will be set to -1.
    ctab : ndarray, shape (n_labels, 5)
        RGBT + label id colortable array.
    names : list of bytes
        The names of the labels. The length of the list is n_labels.
    """
    with open(filepath, "rb") as fobj:
        dt = _ANNOT_DT

        # number of vertices
        vnum = np.fromfile(fobj, dt, 1)[0]

        # vertex ids + annotation values
        data = np.fromfile(fobj, dt, vnum * 2).reshape(vnum, 2)
        labels = data[:, 1]

        # is there a color table?
        ctab_exists = np.fromfile(fobj, dt, 1)[0]
        if not ctab_exists:
            raise Exception('Color table not found in annotation file')

        # in old-format files, the next field will contain the number of
        # entries in the color table. In new-format files, this must be
        # equal to -2
        n_entries = np.fromfile(fobj, dt, 1)[0]

        # We've got an old-format .annot file.
        if n_entries > 0:
            ctab, names = _read_annot_ctab_old_format(fobj, n_entries)
        # We've got a new-format .annot file
        else:
            ctab, names = _read_annot_ctab_new_format(fobj, -n_entries)

    # generate annotation values for each LUT entry
    ctab[:, [4]] = _pack_rgb(ctab[:, :3])

    if not orig_ids:
        ord = np.argsort(ctab[:, -1])
        mask = labels != 0
        labels[~mask] = -1
        labels[mask] = ord[np.searchsorted(ctab[ord, -1], labels[mask])]
    return labels, ctab, names

def _pack_rgb(rgb):
    """Pack an RGB sequence into a single integer.

    Used by :func:`read_annot` and :func:`write_annot` to generate
    "annotation values" for a Freesurfer ``.annot`` file.

    Parameters
    ----------
    rgb : ndarray, shape (n, 3)
        RGB colors

    Returns
    -------
    out : ndarray, shape (n, 1)
        Annotation values for each color.
    """
    bitshifts = 2 ** np.array([[0], [8], [16]], dtype=rgb.dtype)
    return rgb.dot(bitshifts)

def _read_annot_ctab_old_format(fobj, n_entries):
    """Read in an old-style Freesurfer color table from `fobj`.

    Note that the output color table ``ctab`` is in RGBT form, where T
    (transparency) is 255 - alpha.

    This function is used by :func:`read_annot`.

    Parameters
    ----------

    fobj : file-like
        Open file handle to a Freesurfer `.annot` file, with seek point
        at the beginning of the color table data.
    n_entries : int
        Number of entries in the color table.

    Returns
    -------

    ctab : ndarray, shape (n_entries, 5)
        RGBT colortable array - the last column contains all zeros.
    names : list of str
        The names of the labels. The length of the list is n_entries.
    """
    assert hasattr(fobj, 'read')

    dt = _ANNOT_DT
    # orig_tab string length + string
    length = np.fromfile(fobj, dt, 1)[0]
    orig_tab = np.fromfile(fobj, '>c', length)
    orig_tab = orig_tab[:-1]
    names = list()
    ctab = np.zeros((n_entries, 5), dt)
    for i in range(n_entries):
        # structure name length + string
        name_length = np.fromfile(fobj, dt, 1)[0]
        name = np.fromfile(fobj, "|S%d" % name_length, 1)[0]
        names.append(name)
        # read RGBT for this entry
        ctab[i, :4] = np.fromfile(fobj, dt, 4)

    return ctab, names


def _read_annot_ctab_new_format(fobj, ctab_version):
    """Read in a new-style Freesurfer color table from `fobj`.

    Note that the output color table ``ctab`` is in RGBT form, where T
    (transparency) is 255 - alpha.

    This function is used by :func:`read_annot`.

    Parameters
    ----------

    fobj : file-like
        Open file handle to a Freesurfer `.annot` file, with seek point
        at the beginning of the color table data.
    ctab_version : int
        Color table format version - must be equal to 2

    Returns
    -------

    ctab : ndarray, shape (n_labels, 5)
        RGBT colortable array - the last column contains all zeros.
    names : list of str
        The names of the labels. The length of the list is n_labels.
    """
    assert hasattr(fobj, 'read')

    dt = _ANNOT_DT
    # This code works with a file version == 2, nothing else
    if ctab_version != 2:
        raise Exception('Unrecognised .annot file version (%i)', ctab_version)
    # maximum LUT index present in the file
    max_index = np.fromfile(fobj, dt, 1)[0]
    ctab = np.zeros((max_index, 5), dt)
    # orig_tab string length + string
    length = np.fromfile(fobj, dt, 1)[0]
    np.fromfile(fobj, "|S%d" % length, 1)[0]  # Orig table path
    # number of LUT entries present in the file
    entries_to_read = np.fromfile(fobj, dt, 1)[0]
    names = list()
    for _ in range(entries_to_read):
        # index of this entry
        idx = np.fromfile(fobj, dt, 1)[0]
        # structure name length + string
        name_length = np.fromfile(fobj, dt, 1)[0]
        name = np.fromfile(fobj, "|S%d" % name_length, 1)[0]
        names.append(name)
        # RGBT
        ctab[idx, :4] = np.fromfile(fobj, dt, 4)

    return ctab, names
    
def _fread3(fobj):
    """Read a 3-byte int from an open binary file object
    Parameters
    ----------
    fobj : file
        File descriptor
    Returns
    -------
    n : int
        A 3 byte int
    """
    b1, b2, b3 = np.fromfile(fobj, ">u1", 3)
    return (b1 << 16) + (b2 << 8) + b3


def _read_volume_info(fobj):
    """Helper for reading the footer from a surface file."""
    volume_info = OrderedDict()
    head = np.fromfile(fobj, ">i4", 1)
    if not np.array_equal(head, [20]):  # Read two bytes more
        head = np.concatenate([head, np.fromfile(fobj, ">i4", 2)])
        if not np.array_equal(head, [2, 0, 20]) and not np.array_equal(
            head, [2, 1, 20]
        ):
            warnings.warn("Unknown extension code.")
            return volume_info
        head = [2, 0, 20]

    volume_info["head"] = head
    for key in [
        "valid",
        "filename",
        "volume",
        "voxelsize",
        "xras",
        "yras",
        "zras",
        "cras",
    ]:
        pair = fobj.readline().decode("utf-8").split("=")
        if pair[0].strip() != key or len(pair) != 2:
            raise IOError("Error parsing volume info.")
        if key in ("valid", "filename"):
            volume_info[key] = pair[1].strip()
        elif key == "volume":
            volume_info[key] = np.array(pair[1].split()).astype(int)
        else:
            volume_info[key] = np.array(pair[1].split()).astype(float)
    # Ignore the rest
    return volume_info    
    
def read_geometry(filepath, read_metadata=False, read_stamp=False):
    """Read a triangular format Freesurfer surface mesh.
    Parameters
    ----------
    filepath : str
        Path to surface file.
    read_metadata : bool, optional
        If True, read and return metadata as key-value pairs.
        Valid keys:
        * 'head' : array of int
        * 'valid' : str
        * 'filename' : str
        * 'volume' : array of int, shape (3,)
        * 'voxelsize' : array of float, shape (3,)
        * 'xras' : array of float, shape (3,)
        * 'yras' : array of float, shape (3,)
        * 'zras' : array of float, shape (3,)
        * 'cras' : array of float, shape (3,)
    read_stamp : bool, optional
        Return the comment from the file
    Returns
    -------
    coords : numpy array
        nvtx x 3 array of vertex (x, y, z) coordinates.
    faces : numpy array
        nfaces x 3 array of defining mesh triangles.
    volume_info : OrderedDict
        Returned only if `read_metadata` is True.  Key-value pairs found in the
        geometry file.
    create_stamp : str
        Returned only if `read_stamp` is True.  The comment added by the
        program that saved the file.
    """
    volume_info = OrderedDict()

    TRIANGLE_MAGIC = 16777214

    with open(filepath, "rb") as fobj:
        magic = _fread3(fobj)

        if magic == TRIANGLE_MAGIC:  # Triangle file
            create_stamp = fobj.readline().rstrip(b"\n").decode("utf-8")
            test_dev = fobj.peek(1)[:1]
            if test_dev == b"\n":
                fobj.readline()
            vnum = np.fromfile(fobj, ">i4", 1)[0]
            fnum = np.fromfile(fobj, ">i4", 1)[0]
            coords = np.fromfile(fobj, ">f4", vnum * 3).reshape(vnum, 3)
            faces = np.fromfile(fobj, ">i4", fnum * 3).reshape(fnum, 3)

            if read_metadata:
                volume_info = _read_volume_info(fobj)
        else:
            raise ValueError(
                "File does not appear to be a Freesurfer surface (triangle file)"
            )

    coords = coords.astype(float)  # XXX: due to mayavi bug on mac 32bits

    ret = (coords, faces)
    if read_metadata:
        if len(volume_info) == 0:
            warnings.warn("No volume information contained in the file")
        ret += (volume_info,)
    if read_stamp:
        ret += (create_stamp,)

    return ret

def read_label(filepath, read_scalars=False):
    """Load in a Freesurfer .label file.

    Parameters
    ----------
    filepath : str
        Path to label file.
    read_scalars : bool, optional
        If True, read and return scalars associated with each vertex.

    Returns
    -------
    label_array : numpy array
        Array with indices of vertices included in label.
    scalar_array : numpy array (floats)
        Only returned if `read_scalars` is True.  Array of scalar data for each
        vertex.
    """
    label_array = np.loadtxt(filepath, dtype=int, skiprows=2, usecols=[0])
    if read_scalars:
        scalar_array = np.loadtxt(filepath, skiprows=2, usecols=[-1])
        return label_array, scalar_array
    return label_array

def write_label(filepath, vertices, values=None):
    """
    Write Freesurfer label data `values` to filepath `filepath`

    Parameters
    ----------
    filepath : str
        String containing path to label file to be written
    vertices : ndarray, shape (n_vertices, 3)
        Coordinates of each vertex
    values : optional, shape (n_vertices,)
        Array of scalar data for each vertex. The default is None.
    """
    
    if values is not None:
        vector = np.asarray(values)
        vnum = np.prod(vector.shape)
        if vector.shape not in ((vnum,), (vnum, 1), (1, vnum), (vnum, 1, 1)):
            raise ValueError('Invalid shape: argument values must be a vector')
            
    else:
        vector = np.zeros(vnum, dtype=np.float32)
    
    start_line = '#!ascii label  , from subject  vox2ras=TkReg\n'
    magic_number = vnum + 6000
    
    with open(filepath, 'w') as fobj:
        fobj.write(start_line)
        fobj.write(f'{magic_number}\n')
        
        #now write vertex and label array
        label_array = np.vstack((np.array(range(vnum)), vertices.T, vector.T)).T.astype('>f4')
        np.savetxt(fobj, label_array, fmt=['%i','%f','%f','%f','%f'])
        
def read_morph_data(filepath):
    """Read a Freesurfer morphometry data file.

    This function reads in what Freesurfer internally calls "curv" file types,
    (e.g. ?h. curv, ?h.thickness), but as that has the potential to cause
    confusion where "curv" also refers to the surface curvature values,
    we refer to these files as "morphometry" files with PySurfer.

    Parameters
    ----------
    filepath : str
        Path to morphometry file

    Returns
    -------
    curv : numpy array
        Vector representation of surface morpometry values
    """
    with open(filepath, "rb") as fobj:
        magic = _fread3(fobj)
        if magic == 16777215:
            vnum = np.fromfile(fobj, ">i4", 3)[0]
            curv = np.fromfile(fobj, ">f4", vnum)
        else:
            vnum = magic
            _fread3(fobj)
            curv = np.fromfile(fobj, ">i2", vnum) / 100
    return curv

def geodesic_distmat(surf, parc=None, n_jobs=1, use_cholmod=True):
    """
    Compute geodesic distance using the heat diffusion method built into LaPy
        Based on: doi/10.1145/2516971.2516977
        Crane et al., "Geodesics in heat: A new approach to computing distance 
        based on heat flow"
    
    Parameters
    ----------
    surf : Shape class
        Input surface
    n_jobs : int, optional
        Number of workers to use for parallel calls to ._thread_method(),
        default is 1.
    
    Returns
    -------
    D : (N,N) np.ndarray
        Distance matrix of every vertex to every other vertex
    
    """
    surf = nib.load(surf).agg_data()
    tria = TriaMesh(surf[0], surf[1])
    try:
        import_optional_dependency('sksparse')
        use_cholmod=True
    except:
        use_cholmod=False
    
    fem = Solver(tria, lump=True, use_cholmod=use_cholmod)
        
    D = __distance_threading__(tria, fem, n_jobs=n_jobs)
    
    if parc:
        D = np.row_stack([
            D[parc == lab].mean(axis=0) for lab in np.unique(parc)
        ])
        D[np.diag_indices_from(D)] = 0
        D = D[1:, 1:]
    
    return D

def __distance_threading__(tria, fem, n_jobs=1):

    D = np.column_stack(
        Parallel(n_jobs=n_jobs, prefer='threads')(
            delayed(_distance_thread_method)(tria, fem, bvert=bvert) for bvert in range(tria.v.shape[0])
            )
        )
    
    return np.asarray(D.squeeze())

def _distance_thread_method(tria, fem, bvert):    
    u = diffusion(tria, fem, vids=bvert, m=1.0)
    
    tfunc = tria_compute_gradient(tria, u)
    
    X = -tfunc / np.sqrt((tfunc**2).sum(1))[:,np.newaxis]
    X = np.nan_to_num(X)
    
    b0 = tria_compute_divergence(tria, X)
    
    if fem.use_cholmod:
        chol = cholesky(-fem.stiffness)
        d = chol(b0)
    else:
        A = fem.stiffness
        H = -A
        lu = splu(H)
        d = lu.solve(b0.astype(np.float32))
    
    d = d - min(d)
        
    return d

def diffusion(geometry, fem, vids, m=1.0, use_cholmod=True):
    """
    Computes heat diffusion from initial vertices in vids using
    backward Euler solution for time t [MO2]:
    
      t = m * avg_edge_length^2
    
    Parameters
    ----------
      geometry      TriaMesh or TetMesh, on which to run diffusion
      vids          vertex index or indices where initial heat is applied
      m             factor (default 1) to compute time of heat evolution:
                    t = m * avg_edge_length^2
      use_cholmod   (default True), if Cholmod is not found
                    revert to LU decomposition (slower)
    
    Returns
    -------
      vfunc         heat diffusion at vertices
    """
    sksparse = import_optional_dependency("sksparse", raise_error=use_cholmod)
    
    nv = len(geometry.v)
    fem = fem
    # time of heat evolution:
    t = m * geometry.avg_edge_length() ** 2
    # backward Euler matrix:
    hmat = fem.mass + t * fem.stiffness
    # set initial heat
    b0 = np.zeros((nv,))
    b0[np.array(vids)] = 1.0
    # solve H x = b0
    #print("Matrix Format now:  " + hmat.getformat())
    if sksparse is not None:
        print("Solver: Cholesky decomposition from scikit-sparse cholmod ...")
        chol = sksparse.cholmod.cholesky(hmat)
        vfunc = chol(b0)
    else:
        from scipy.sparse.linalg import splu
    
        print("Solver: spsolve (LU decomposition) ...")
        lu = splu(hmat)
        vfunc = lu.solve(np.float32(b0))
    return vfunc

"""Handle optional dependency imports.

Inspired from pandas: https://pandas.pydata.org/
"""

import importlib

# A mapping from import name to package name (on PyPI) when the package name
# is different.
INSTALL_MAPPING = {
    "sksparse": "scikit-sparse",
}


def import_optional_dependency(
    name: str,
    extra: str = "",
    raise_error: bool = True,
):
    """Import an optional dependency.

    By default, if a dependency is missing an ImportError with a nice message
    will be raised.

    Parameters
    ----------
    name : str
        The module name.
    extra : str, default=""
        Additional text to include in the ImportError message.
    raise_error : bool, default=True
        What to do when a dependency is not found.
        * True : Raise an ImportError.
        * False: Return None.

    Returns
    -------
    module : Optional[ModuleType]
        The imported module when found.
        None is returned when the package is not found and raise_error is
        False.

    Raises
    -------
    ImportError
        dependency not found; see raise_error
    """

    package_name = INSTALL_MAPPING.get(name)
    install_name = package_name if package_name is not None else name

    try:
        module = importlib.import_module(name)
    except ImportError:
        if raise_error:
            raise ImportError(
                f"Missing optional dependency '{install_name}'. {extra} "
                f"Use pip or conda to install {install_name}."
            )
        else:
            return None

    return module


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Data fetchers
"""

import os.path as op
from pathlib import Path
import shutil

import numpy as np
from neuromaps import datasets, images
import nibabel as nib
from .utils import (get_data_dir, _groupby_match,
                    get_dataset_info, _match_files)
from eigenstrapping import dataio

from nilearn.datasets.utils import _fetch_file

def load_surface_examples(data_dir=None, with_surface=False):
    """
    Downloads all surface files for fsaverage5 space

    Parameters
    ----------
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'EIGEN_DATA'; if that is not set, will use
        `~/eigenstrapping-data` instead. Default: None
    with_surface : str, optional
        Flag whether to return the surfaces or not. Default False.

    Returns
    -------
    tuple of ndarrays
        Maps, emodes, and evals for left and right hemispheres.

    """

    space = 'fsaverage'
    version = 'fsaverage5'
    den = '10k'
        
    local_path = op.dirname(__file__)
    
    data_fname = '{}_{}.gii'
    mask_fname = '{}_{}.txt'
    
    data_pth = op.join(local_path, 'brainmaps', data_fname)
    mask_pth = op.join(local_path, 'masks', mask_fname)
    
    surfs = [None] * 2
    data = [None] * 2
    emodes = [None] * 2
    evals = [None] * 2
    masks = [None] * 2
    for i, side in enumerate(['lh', 'rh']):
        surfs[i] = fetch_data(name='surfaces', space=space, den=den,
                              hemi=side)
        data[i] = nib.load(data_pth.format(version, side)).agg_data()
        emodes_file = fetch_data(name='eigenmodes', space=space, den=den, hemi=side,
                   format='emodes')
        emodes[i] = np.loadtxt(emodes_file)
        evals_file = fetch_data(name='eigenmodes', space=space, den=den, hemi=side,
                   format='evals')
        evals[i] = np.loadtxt(evals_file)
        masks[i] = np.loadtxt(mask_pth.format(version, side)).astype(np.bool_)
        data[i][~masks[i]] = np.nan
    
    if with_surface:
        return surfs[0], surfs[1], data[0], data[1], emodes[0], emodes[1], evals[0], evals[1]
    
    return data[0], data[1], emodes[0], emodes[1], evals[0], evals[1]

def load_genepc(join=False):
    """
    Download and return the Allen Human Brain Atlas gene PC1 in
    fsaverage5 space.

    Parameters
    ----------
    join : bool, optional
        If True, return both hemispheres in one array. If False, return
        two hemispheres separately. Default is False.

    Returns
    -------
    tuple of ndarrays of shape (10242,) (10242,)
        if join is True then returns ndarray of shape (20484,)
        Gene PC1 data in fsaverage5 space
    """
    genepc = datasets.fetch_annotation(desc='genepc1', return_single=True)
    data = images.load_data(genepc)
    
    if join:
        return data
    
    data_lh, data_rh = data.reshape((2, 10242))
    
    return data_lh, data_rh

def load_distmat(space='fsaverage', den='10k', data_dir=None, hemi='lh', parcellated=False, sort=True):
    """
    Downloads geodesic distance matrices. If the dense distance matrix is retrieved, 
    there is an option to return the sorted distance matrix and index memmapped, ala
    variogram calculation in ``brainsmash`` in `sort=True`. If `sort=False`,
    the unsorted distance matrix (as per normal) is returned.

    Parameters
    ----------
    space : str, optional
        Which space of the files to get. Default is 'fsaverage5'
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'E_DATA'; if that is not set, will use
        `~/e-data` instead. Default: None
    hemi : str, optional
        Which hemisphere to load data for. Default 'lh'
    parcellated : bool, optional
        If True, return parcellated format of distance matrix. If False, return
        sorted dense distance matrix.
    sort : bool, optional
        If True, return sorted dense distance matrix and index memory-mapped using
        `.utils.txt2memmap()` for variogram calculation. Does nothing if
        `parcellation` is passed.
        
    Returns
    -------
    if parcellated:
        ndarray
            Parcellated distance matrix
    else:
        if sort:
            tuple of ndarrays
                Sorted distance matrix and index for hemisphere in 'lh'
        else:
            ndarray
            Dense geodesic distance matrix

    """
    if parcellated:
        distmat = fetch_data(name='distmat', space=space, den=den,
                             hemi=hemi, format='parc')
        return distmat
    
    distmat = fetch_data(name='distmat', space=space, den=den,
                         hemi=hemi, format='dense')
    
    outfiles = txt2memmap(distmat, get_data_dir(data_dir))
    D = load_memmap(outfiles['distmat'])
    index = load_memmap(outfiles['index'])
        
    return D, index

def load_subcort(structure='thalamus', data_dir=None):
    """
    Loads all the files for the subcortical structures used in the paper -
    cortical-subcortical gradients and the subcortical ROIs from the 25% 
    probability Harvard-Oxford atlas.

    Parameters
    ----------
    structure : str, optional
        Which subcortical structure to return. Default is 'thalamus'.

    Returns
    -------
    data : nib.Nifti1Image
        Subcortical data of cortico-subcortical gradients in MNI152 space.
    mask : nib.Nifti1Image
        Mask image of subcortical structure in MNI152 space (i.e., 1 inside 
        structure and 0 outside structure)

    """
    root_pth = op.dirname(__file__)
    data = nib.load(op.join(root_pth, f'brainmaps/{structure}.txt'))
    mask = nib.load(op.join(root_pth, f'masks/{structure}.txt'))
    
    return data, mask

def load_native(data_dir=None):
    """
    Download HCP native surface and data for demonstration purposes.

    Parameters
    ----------
    data_dir : str, optional
        Path to use . The default is Npne.

    Returns
    -------
    native : dict
        Dictionary of filenames 

    """
    native_surface = fetch_data(name='surfaces', space='native', data_dir=data_dir)
    native_data = fetch_data(name='brainmaps', space='native', data_dir=data_dir)
    
    return {'surface' : native_surface[0], 'data' : native_data[0]}

def fetch_data(*, name=None, space=None, den=None, res=None, hemi=None,
                   tags=None, format=None, data_dir=None, verbose=1):
    """
    Downloads files for brain surfaces and eigenmodes matching requested variables

    Parameters
    ----------
    name, space, den, res, hemi, tags, format : str or list-of-str
        Values on which to match surfaces. If not specified surfaces with
        any value for the relevant key will be matched. Default: None
    data_dir : str, optional
        Path to use as data directory. If not specified, will
        use `~/e-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    Returns
    -------
    data : dict
        Dictionary of downloaded annotations where dictionary keys are tuples
        (space, den/res) and values are lists of corresponding
        filenames
    """

    # check input parameters to ensure we're fetching _something_
    supplied = False
    for val in (space, den, res, hemi, tags, format):
        if val is not None:
            supplied = True
            break
    if not supplied:
        raise ValueError('Must provide at least one parameters on which to '
                         'match files. If you want to fetch all '
                         'annotations set any of the parameters to "all".')

    # get info on datasets we need to fetch
    data_dir = get_data_dir(data_dir=data_dir)
    info = _match_files(get_dataset_info(name),
                        space=space, den=den, res=res,
                        hemi=hemi, tags=tags, format=format)
    if verbose > 1:
        print(f'Identified {len(info)} datsets matching specified parameters')

    # TODO: current work-around to handle that _fetch_files() does not support
    # session instances. hopefully a future version will and we can just use
    # that function to handle this instead of calling _fetch_file() directly
    data = []
    for dset in info:
        fn = Path(data_dir) / dset['rel_path'] / dset['fname']
        if not fn.exists():
            dl_file = _fetch_file(dset['url'], str(fn.parent), verbose=verbose,
                                  md5sum=dset['checksum'])
            shutil.move(dl_file, fn)
        data.append(str(fn))
    
    if len(data) == 1:
        return data[0]
    
    return _groupby_match(data)

def txt2memmap(dist_file, output_dir, maskfile=None, delimiter=' '):
    """
    Export distance matrix to memory-mapped array.

    Parameters
    ----------
    dist_file : filename
        Path to `delimiter`-separated distance matrix file
    output_dir : filename
        Path to directory in which output files will be written
    maskfile : filename or np.ndarray or None, default None
        Path to a neuroimaging/txt file containing a mask, or a mask
        represented as a numpy array. Mask scalars are cast to boolean, and
        all elements not equal to zero will be masked.
    delimiter : str
        Delimiting character in `dist_file`

    Returns
    -------
    dict
        Keys are 'D' and 'index'; values are absolute paths to the
        corresponding binary files on disk.

    Notes
    -----
    Each row of the distance matrix is sorted before writing to file. Thus, a
    second mem-mapped array is necessary, the i-th row of which contains
    argsort(d[i]).
    If `maskfile` is not None, a binary mask.txt file will also be written to
    the output directory.

    Raises
    ------
    IOError : `output_dir` doesn't exist
    ValueError : Mask image and distance matrix have inconsistent sizes

    """

    nlines = count_lines(dist_file)
    if not op.exists(output_dir):
        raise IOError("Output directory does not exist: {}".format(output_dir))

    # Load mask if one was provided
    if maskfile is not None:
        mask = dataio(maskfile).astype(bool)
        if mask.size != nlines:
            e = "Incompatible input sizes\n"
            e += "{} rows in {}\n".format(nlines, dist_file)
            e += "{} elements in {}".format(mask.size, maskfile)
            raise ValueError(e)
        mask_fileout = op.join(output_dir, "mask.txt")
        np.savetxt(  # Write to text file
            fname=mask_fileout, X=mask.astype(int), fmt="%i", delimiter=',')
        nv = int((~mask).sum())  # number of non-masked elements
        idx = np.arange(nlines)[~mask]  # indices of non-masked elements
    else:
        nv = nlines
        idx = np.arange(nlines)

    # Build memory-mapped arrays
    with open(dist_file, 'r') as fp:

        npydfile = op.join(output_dir, "distmat.npy")
        npyifile = op.join(output_dir, "index.npy")
        fpd = np.lib.format.open_memmap(
            npydfile, mode='w+', dtype=np.float32, shape=(nv, nv))
        fpi = np.lib.format.open_memmap(
            npyifile, mode='w+', dtype=np.int32, shape=(nv, nv))

        ifp = 0  # Build memory-mapped arrays one row of distances at a time
        for il, l in enumerate(fp):  # Loop over lines of file
            if il not in idx:  # Keep only CIFTI vertices
                continue
            else:
                line = l.rstrip()
                if line:
                    data = np.array(line.split(delimiter), dtype=np.float32)
                    if data.size != nlines:
                        raise RuntimeError(
                            "Distance matrix is not square: {}".format(
                                dist_file))
                    d = data[idx]
                    sort_idx = np.argsort(d)
                    fpd[ifp, :] = d[sort_idx]  # sorted row of distances
                    fpi[ifp, :] = sort_idx  # sort indexes
                    ifp += 1
        del fpd  # Flush memory changes to disk
        del fpi

    return {'distmat': npydfile, 'index': npyifile}  # Return filenames


def load_memmap(filename):
    """
    Load a memory-mapped array.

    Parameters
    ----------
    filename : str
        path to memory-mapped array saved as npy file

    Returns
    -------
    np.memmap

    """
    return np.load(filename, mmap_mode='r')

def count_lines(filename):
    """
    Count number of lines in a file.

    Parameters
    ----------
    filename : filename

    Returns
    -------
    int
        number of lines in file

    """
    with open(filename, 'rb') as f:
        lines = 0
        buf_size = 1024 * 1024
        read_f = f.raw.read
        buf = read_f(buf_size)
        while buf:
            lines += buf.count(b'\n')
            buf = read_f(buf_size)
        return lines


from .base import (SurfaceEigenstrapping, VolumetricEigenstrapping,
                   compute_psd)

__all__ = ['SurfaceEigenstrapping',
           'VolumetricEigenstrapping',
           'compute_psd'
           ]


# -*- coding: utf-8 -*-

import os
from pathlib import Path
import warnings

import nibabel as nib
import numpy as np
from scipy import ndimage
import seaborn as sns
import tqdm

from netneurotools import datasets as nndata
from netneurotools.freesurfer import _decode_list

DROP = [  # regions that should always be dropped from analyses
    'lh_unknown', 'rh_unknown',
    'lh_corpuscallosum', 'rh_corpuscallosum',
    'lh_Background+FreeSurfer_Defined_Medial_Wall',
    'rh_Background+FreeSurfer_Defined_Medial_Wall',
]


def rgb255(x):
    return np.around(np.asarray(x) * 255)


REDS = rgb255(sns.color_palette('Reds', 7, desat=0.7)[-5:])
BLUES = rgb255(sns.color_palette('Blues', 4, desat=0.5)[-3:])
PURPLES = rgb255(sns.color_palette('Purples', 5, desat=0.8))[[2, 4]]
SPATHUES = list(np.r_[PURPLES, REDS, BLUES] / 255)
PARCHUES = list(np.array([[26, 146, 0], [222, 57, 90], [131, 131, 131]]) / 255)


def pathify(path):
    """
    Convenience function for coercing a potential pathlike to a Path object

    Parameter
    ---------
    path : str or os.PathLike
        Path to be checked for coercion to pathlib.Path object

    Returns
    -------
    path : pathlib.Path
    """

    if isinstance(path, (str, os.PathLike)):
        path = Path(path)
    return path


def get_cammoun_schaefer(vers='fsaverage5', data_dir=None, networks='7'):
    """
    Returns Cammoun 2012 and Schaefer 2018 atlases as dictionary

    Parameters
    ----------
    vers : str, optional
        Which version of the atlases to get. Default: 'fsaverage5'
    data_dir : str or os.PathLike, optional
        Data directory where downloaded atlases should be stored. If not
        specified will default to $NNT_DATA or ~/nnt-data
    networks : {'7', '17'}, optional
        Which networks to get for Schaefer 2018 atlas. Default: '7'

    Returns
    -------
    atlases : dict
        Where keys are 'atl-cammoun2012' and 'atl-schaefer2018'
    """
    cammoun = nndata.fetch_cammoun2012(vers, data_dir=data_dir)
    schaefer = nndata.fetch_schaefer2018('fsaverage5', data_dir=data_dir)
    schaefer = {k: schaefer.get(k) for k in schaefer.keys()
                if 'Parcels7Networks' in k}

    return {'atl-cammoun2012': cammoun, 'atl-schaefer2018': schaefer}


def save_dir(fname, data, overwrite=True):
    """
    Saves `data` to `fname`, creating any necessary intermediate directories

    Parameters
    ----------
    fname : str or os.PathLike
        Output filename for `data`
    data : array_like
        Data to be saved to disk
    """

    fname = Path(fname).resolve()
    fname.parent.mkdir(parents=True, exist_ok=True)
    fmt = '%.10f' if data.dtype.kind == 'f' else '%d'
    if fname.exists() and not overwrite:
        warnings.warn(f'{fname} already exists; not overwriting')
        return
    np.savetxt(fname, data, delimiter=',', fmt=fmt)


def parcellate(mgh, annot, drop=None):
    """
    Parcellates surface `mgh` file with `annot`

    Parameters
    ----------
    mgh : str or os.PathLike
        Surface data to be parcellated
    annot : str or os.PathLike
        FreeSurfer-style annotation file matching `mgh`
    drop : list-of-str, optional
        List of parcels to `drop` from parcellated data. Default: None

    Returns
    -------
    data : numpy.ndarray
        Parcellated data from `mgh`
    """

    try:
        img = nib.load(mgh).dataobj
    except ValueError:
        img = mgh

    labels, ctab, names = nib.freesurfer.read_annot(annot)
    data = ndimage.mean(np.squeeze(img), labels, np.unique(labels))

    if drop is not None:
        if isinstance(drop, str):
            drop = [drop]
        drop, names = _decode_list(drop), _decode_list(names)
        drop = np.intersect1d(names, drop)
        data = np.delete(data, [names.index(f) for f in drop])

    return data


def get_names(*, lh, rh):
    """
    Gets parcel labels in `{l,r}h` files and prepends hemisphere designation

    Parameters
    ----------
    {l,r}h : str or os.PathLike
        FreeSurfer-style annotation files for the left / right hemisphere

    Returns
    -------
    names : list
        List of parcel names with 'lh_' or 'rh_' prepended to names
    """

    names = []
    for fn, hemi in ((lh, 'lh'), (rh, 'rh')):
        n = nib.freesurfer.read_annot(fn)[-1]
        names += [
            f'{hemi}_{fn.decode()}' if hasattr(fn, 'decode') else fn
            for fn in n
        ]
    return names


def trange(n_iter, verbose=True, **kwargs):
    """
    Wrapper for :obj:`tqdm.trange` with some default options set

    Parameters
    ----------
    n_iter : int
        Number of iterations for progress bar
    verbose : bool, optional
        Whether to return an :obj:`tqdm.tqdm` progress bar instead of a range
        generator. Default: True
    kwargs
        Key-value arguments provided to :func:`tqdm.trange`

    Returns
    -------
    progbar : :obj:`tqdm.tqdm`
    """

    form = ('{desc}: {percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt}'
            ' | {elapsed}<{remaining}')
    defaults = dict(ascii=True, leave=False, bar_format=form)
    defaults.update(kwargs)

    return tqdm.trange(n_iter, disable=not verbose, **defaults)


def yield_data_dist(dist_dir, atlas, scale, data, medial=False, inverse=True):
    """
    Yields hemisphere-specific data and distance matrices

    Parameters
    ----------
    dist_dir : str or os.PathLike
        Path to directory where distance matrices are stored
    atlas : {'atl-cammoun2012', 'atl-schaefer2018'}, str
        Name of atlas for which to load data
    scale : str
        Scale of atlas to use
    data : pandas.DataFrame
        Dataframe where index has labels indicating to which hemisphere each
        value belongs
    medial : bool, optional
        Whether to load distance matrix with medial wall travel. Default: False
    inverse : bool, optional
        Whether to return inverse distance matrix. Default: True

    Yields
    -------
    hemidata : (N,) numpy.ndarray
        Data for {left, right} hemisphere
    dist : (N, N) numpy.ndarray
        Distance matrix for {left, right} hemisphere
    index : (N,) list
        Indices of provided `data` that yielded `hemidata`
    """

    medial = ['nomedial', 'medial'][medial]

    for n, hemi in enumerate(('lh', 'rh')):
        # load relevant distance matrix
        fn = pathify(dist_dir) / atlas / medial / f'{scale}_{hemi}_dist.csv'
        npy = fn.with_suffix('.npy')
        if npy.exists():
            dist = np.load(npy, allow_pickle=False, mmap_mode='c')
        else:
            dist = np.loadtxt(fn, delimiter=',')
            np.save(npy, dist, allow_pickle=False)

        if inverse:
            np.fill_diagonal(dist, 1)
            dist **= -1

        # get indices of data that correspond to relevant `hemi` and subset.
        # if data is not a pandas DataFrame with hemisphere information assume
        # we can split the data equally in 2
        try:
            idx = [n for n, f in enumerate(data.index)if f.startswith(hemi)]
            hemidata = np.squeeze(np.asarray(data.iloc[idx]))
        except AttributeError:
            idx = np.arange(n * (len(data) // 2), (n + 1) * (len(data) // 2))
            hemidata = np.squeeze(data[idx])

        yield hemidata, dist, np.asarray(idx)


def drop_unknown(data, drop=DROP):
    """
    Removes rows from `data` corresponding to entries in `drop`

    If there is no overlap between `data.index` and `drop` then nothing is done

    Parameters
    ----------
    data : pandas.DataFrame
        Data from which `drop` should be removed
    drop : array_like
        Indices or index names of rows that should be removed from `data`. If
        not supplied then `parspin.utils.DROP` is used. Default: None

    Returns
    -------
    data : pandas.DataFrame
        Provided dataframe with `drop` rows removed
    """

    todrop = np.array(drop)[np.isin(drop, data.index)]
    if len(todrop) > 0:
        data = data.drop(todrop, axis=0)

    return data


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Plotting subroutines
"""

from nilearn import plotting
import matplotlib.pyplot as plt
from matplotlib import gridspec, cm
import numpy as np
import os
from neuromaps.datasets import fetch_atlas

surfaces = ['fsaverage', 'fsLR']

def vrange(scalars, pv=5.):
    # check if nans
    mask = np.isnan(scalars)
    
    # get visual range
    return np.percentile(scalars[~mask], pv), np.percentile(scalars[~mask], 100-pv)

def csplot(scalars, surface='fsLR', hemisphere='left', orientation='landscape', 
           vrange=(-4,4), cmap='viridis'):
    """
    Automated plotting of cortical surfaces. Generates a figure of
    a scalar brainmap `y` and returns `fig` and `ax` handles for
    further use (if needed). Suitable for use with data registered to
    CIFTI 32k space only. Plots data on fs_LR_32k freesurfer-type 
    "midthickness" surface.

    Parameters
    ----------
    scalars : np.ndarray of shape=(32492,)
        Brainmap to visualize registered to standard fs_LR_32k CIFTI space.
    hemisphere : str, optional
        Hemisphere to plot. The default is 'left'.
    orientation : str, optional
        Orientation of plot, 'landscape' or 'portrait'. Default is 'landscape'.
    vrange : tuple, optional
        Maximum range of values to plot. Values higher or lower are truncated.
        Default is (-4, 4).
    cmap : str, optional
        matplotlib.cm color types. The default is 'viridis'.

    Returns
    -------
    fig : matplotlib.figure handle
    ax : matplotlib.ax handle
    
    Notes
    -----
    If there are NaNs present in the brainmap `y`, these are masked out
    of the plotting. See ``nilearn.plotting``

    """
    if surface not in surfaces:
        raise ValueError('chosen surface must be "fsLR" or "fsaverage5"')
        
    if surface == 'fsaverage':
        surf = 'inflated'
        den = '10k'
    else:
        surf = 'midthickness'
        den = '32k'
        
    # make figure
    if orientation == 'landscape':
        fig = plt.figure(figsize=(9, 6), constrained_layout=False)
        grid = gridspec.GridSpec(1, 2)
        rot = 'vertical'
        axes = [1.01, 0.28, 0.03, 0.4]
    elif orientation == 'portrait':
        fig = plt.figure(figsize=(6, 9), constrained_layout=False)
        grid = gridspec.GridSpec(2, 1)
        rot = 'horizontal'
        axes = [1.01, 0.28, 0.03, 0.4]
        
    if hemisphere == 'left':
        mesh = fetch_atlas(atlas=f'{surface}', density=den)[surf][0]
        bg_map = fetch_atlas(atlas=f'{surface}', density=den)['vaavg'][0]
    elif hemisphere == 'right':
        mesh = fetch_atlas(atlas=f'{surface}', density=den)[surf][1]
        bg_map = fetch_atlas(atlas=f'{surface}', density=den)['vaavg'][1]
    else:
        raise ValueError('Hemisphere can only be "left" or "right"')
    
    # get colormap
    cmap = plt.get_cmap(cmap)
    vmin, vmax = vrange
    view = ['lateral', 'medial']
        
    # plot surface
    for idx in range(2):
        ax = fig.add_subplot(grid[idx], projection='3d')
        plotting.plot_surf(mesh, surf_map=scalars, hemi=hemisphere, 
                           view=view[idx], vmin=vmin, vmax=vmax, 
                           colorbar=False, cmap=cmap, axes=ax,
                           bg_map=bg_map)
        ax.dist = 9
    
    # colorbar
    cax = plt.axes(axes)
    cbar = fig.colorbar(cm.ScalarMappable(norm=None, cmap=cmap), cax=cax, 
                        orientation=rot)
    cbar.set_ticks([])
    cbar.ax.set_title(f'{vmax:.2f}', fontdict={'fontsize':20}, pad=20)
    cbar.ax.set_xlabel(f'{vmin:.2f}', fontdict={'fontsize':20}, labelpad=20)
    
    return fig, ax
    
def flatplot(scalars, hemisphere='left', vrange=(-4,4), cmap='viridis', colorbar=False):
    """
    Automated plotting of cortical surfaces. Generates a figure of
    a scalar brainmap `y` and returns `fig` and `ax` handles for
    further use (if needed). Plots on the fs_LR_32k freesurfer-type "flat" surface.
    Suitable for use with data registered to CIFTI 32k space only.

    Parameters
    ----------
    scalars : np.ndarray of shape=(32492,)
        Brainmap to visualize registered to standard fs_LR_32k CIFTI space.
    hemisphere : str, optional
        Hemisphere to plot. The default is 'left'.
    vrange : tuple, optional
        Maximum range of values to plot. Values higher or lower are truncated.
        Default is (-4, 4).
    cmap : str, optional
        matplotlib.cm color types. The default is 'viridis'.
    colorbar : bool, optional
        If True, plots the colorbar.

    Returns
    -------
    fig : matplotlib.figure handle
    ax : matplotlib.ax handle
    
    Notes
    -----
    If there are NaNs present in the brainmap `y`, these are masked out
    of the plotting. See ``nilearn.plotting``

    """
    # make figure
    fig = plt.figure(figsize=(6, 6), constrained_layout=False)
        
    if hemisphere == 'left':
        mesh = os.path.join(os.path.split(__file__)[0], 'datasets', 'surfaces', 'standard', 'fs_LR.32k.L.flat.surf.gii')
    elif hemisphere == 'right':
        mesh = os.path.join(os.path.split(__file__)[0], 'datasets', 'surfaces', 'standard', 'fs_LR.32k.R.flat.surf.gii')
    else:
        raise ValueError('Hemisphere can only be "left" or "right"')
    
    # get colormap
    cmap = plt.get_cmap(cmap)
    vmin, vmax = vrange
        
    # plot surface
    ax = fig.add_subplot(projection='3d')
    plotting.plot_surf(mesh, surf_map=scalars, hemi=hemisphere, 
                       view='lateral', vmin=vmin, vmax=vmax, 
                       colorbar=False, cmap=cmap, axes=ax)
    ax.dist = 8
    
    # colorbar
    if colorbar:
        cax = plt.axes([1.04, 0.2, 0.03, 0.6])
        cbar = fig.colorbar(cm.ScalarMappable(norm=None, cmap=cmap), cax=cax)
        cbar.set_ticks([])
        cbar.ax.set_title(f'{vmax:.2f}', fontdict={'fontsize':15}, pad=20)
        cbar.ax.set_xlabel(f'{vmin:.2f}', fontdict={'fontsize':15}, labelpad=20)
    
    return fig, ax



"""
Generate random rotation matrices of arbitrary size and rotate matrices
"""
import numpy as np
from numpy.linalg import qr
from sklearn.utils.validation import check_random_state

modes = np

def direct_method(n, seed=None):
    rs = check_random_state(seed)
    # 1. Draw n independent random normal N(0, 1) variates
    v = rs.normal(size=n)
    
    # 2. Normalize
    x = v / np.linalg.norm(v)
    
    # 3. Treat the vector as a single column matrix
    X = x[:, np.newaxis]
    
    # 4. Apply QR decomposition
    Q, _ = qr(X)
    
    return Q

def indirect_method(n, seed=None):
    
    rs = check_random_state(seed)
    # 1. Generate matrix A
    A = rs.normal(size=(n, n))
    
    # 2. Compute the QR decomposition
    Q, R = qr(A)
    
    # 3. Check the determinant of Q
    if np.linalg.det(Q) <= 0:
        A[:, 0] = -A[:, 0]  # Flip the sign of the first column
        Q, R = qr(A)
    
    return Q

def rotate_matrix(M, method='indirect'):
    """
    Rotate an (n/m)-by-n matrix of arbitrary length n by the two methods 
    as outlined in [1].

    Parameters
    ----------
    M : 2D np.ndarray
        Input matrix
    method : str, optional
        Which method to use. Refers to the nomenclature in [1], where
        'indirect' refers to the Householder QR decomposition method [2],
        while 'direct' refers to the method of selecting random points
        on the unit n-sphere directly as in [3]. The default is 'indirect'.

    Returns
    -------
    X_rotated : TYPE
        DESCRIPTION.
        
    References
    ----------
    [1] Blaser R, Fryzlewicz P. Random Rotation Ensembles. Journal of 
        Machine Learning Research. 2016;17(4):1-26.
    [2] Householder A. S. Unitary triangularization of a nonsymmetric 
        matrix. Journal of the ACM, 5:339–342, 1958.
    [3] Knuth D. E. Art of computer programming, volume 2: 
        seminumerical algorithms. Addison-Wesley Professional, Reading, 
        Mass, 3 edition edition, November 1997.

    """
    n = M.shape[1]
    if method == 'indirect':
        rot = indirect_method(n)
        M_rotated = np.dot(M, rot)
    
    elif method == 'direct':
        rot = direct_method(n)
        M_rotated = np.dot(M, rot)
    
    else:
        raise ValueError("Method must be one of 'indirect' or 'direct'")
        
    return M_rotated
    
class RotationMatrix:
    """
    Support class for saving, loading and generating random rotation matrices
    """
    def __init__(self, filename):
        self.filename = filename
        self.data = None
        
    def load_single_matrix(self, l, index):
        """
        Load a single matrix for a given group and index
        """
        n = 2 * l + 1
        num_elements = n * n
        
        # compute starting position
        offset = sum(10 * (2 * i + 1) ** 2 for i in range(1, l)) + index * num_elements
        
        # if data is None, memory-map the entire file
        if self.data is None:
            self.data = np.memmap(self.filename, dtype='float64', mode='r')
            
        matrix_data = self.data[offset:offset+num_elements]
        return np.array(matrix_data).reshape((n, n))
    
    
    def generate_rotation_matrices(self, n, num_matrices=10):
        return np.array([indirect_method(n) for _ in range(num_matrices)])
    
    def generate_list(self, max_l, num_matrices=10):
        return [self.generate_rotation_matrices(2 * l + 1, num_matrices) for l in range(1, max_l + 1)]

    def save_structured_matrices(self, all_rotations_list, filename):
        flat = [matrices.flatten() for matrices in all_rotations_list]
        con = np.concatenate(flat)
        
        np.save(filename, con)
    
    def get_random_rotation(self, l):
        """
        Get random rotation matrix from large array.

        Parameters
        ----------
        l : int
            Group parameter l > 0

        Returns
        -------
        np.ndarray of shape (n, n)
            Random rotation matrix where n = 2 * l + 1

        """
        if l <= 0:
            raise ValueError('Group parameter must be greater than zero, got {}'.format(l))
            
        random_index = np.random.randint(0, 1000)
        return self.load_single_matrix(l, random_index)

    
    
    
    
    
    
    
    
    
    